/*
 * P E R U N -- a physically based lightning generator
 * Master's thesis
 * Department of Computer Graphics and Interaction
 * Faculty of Electrical Engineering
 * Czech Technical University in Prague.
 *
 * Supervisor: Ing. Jaroslav Sloup
 *
 * Distributed under the GNU General Public License v3.0
 * */

/**
 * @file
 * @author Richard Dobrichovsky <dobrichovsky@gmail.com>
 * @brief An implementation of Conjugate Gradient Solvers.
 * 
 * An implementation of a Conjugate Gradient Solver for use in the Perun project. This implementation uses the Eigen library, licenced under the MPL2.
 * The class includes 4 different conjugate gradient method implementations: Eigen on a regular grid, Eigen on an adaptive grid, own implementation based on
 * Shewchuck's paper "An Introduction to the Conjugate Gradient Method Without the Agonizing Pain", page 32, and finally a handler to a CUDA implementation
 * that can be found in the CudaSolver class.
 * 
 * */

#include "cg_solver.h"

/**
 * Adaptive grid version constructor.
 * @param width Simulation grid width.
 * @param height Simulation grid height.
 * @param quadtree Pointer to quadtree structure.
 * @param grown Pointer to set of growth sites.
 * @param grown_last Pointer to set of growth sites in previous frame.
 * @param dest Pointer to set of destination sites.
 * @param potential Pointer to array of potential in grid to write the result to.
 * @param max_iter Maximum number of iterations.
 * @param tolerance Tolerance of solver.
 * @param solverType Solver type: SOLVER_EIGEN_REGULAR, SOLVER_EIGEN, SOLVER_PROG2 or SOLVER_CUDA -- 0, 1, 2 or 3 respectively.
 * */
CGSolver::CGSolver(int width, int height, QuadTree * quadtree, std::unordered_set<int> * grown, std::unordered_set<int> * grown_last, std::unordered_set<int> * dest, double * potential, int max_iter, double tolerance, int solverType)
{
	this->width = width;
	this->height = height;
	this->adaptive = true;
	this->solverType = solverType;

	if(this->solverType == SOLVER_CUDA)
	{
		int deviceCount = 0;
		cudaGetDeviceCount(&deviceCount);
		if(deviceCount == 0)
		{
			std::cout << "CUDA not available, using PROG2 solver.\n";
			this->solverType = SOLVER_PROG2;
		}
	}

	this->grown = grown;
	this->grown_last = grown_last;
	this->dest = dest;
	this->potential = potential;
	this->max_iter = max_iter;
	this->tolerance = tolerance;
	this->quadtree = quadtree;

	this->x = Eigen::VectorXd::Zero(this->width * this->height);
	this->x_old = Eigen::VectorXd::Zero(this->width * this->height);
	this->b = Eigen::VectorXd::Zero(this->width * this->height);
	if(this->solverType == SOLVER_PROG2)
	{
		this->r = Eigen::VectorXd::Zero(this->width * this->height);
		this->r_old = Eigen::VectorXd::Zero(this->width * this->height);
		this->d = Eigen::VectorXd::Zero(this->width * this->height);
		this->Ad = Eigen::VectorXd::Zero(this->width * this->height);	
	}
	else if(this->solverType == SOLVER_EIGEN)
	{
		this->A = Eigen::SparseMatrix<double>(this->width * this->height, this->width * this->height);
		this->A.reserve(Eigen::VectorXi::Constant(this->width*this->height,9));
		this->cg.setMaxIterations(max_iter);
		this->cg.setTolerance(tolerance);	
	}
	else if(this->solverType == SOLVER_CUDA)
	{		
		this->cudaSolver = new CudaSolver(this->tolerance, this->max_iter, this->width, this->height);
	}
	this->rowBeginings.reserve(this->width*this->height);
	for(int i = 0; i<this->width*this->height; ++i)
	{
		this->rowBeginings.push_back(std::list<T>::iterator(this->tripletList.end()));
	}
}

/**
 * Regular grid version constructor. Only uses the Eigen solver.
 * @param width Simulation grid width.
 * @param height Simulation grid height.
 * @param grown Pointer to set of growth sites.
 * @param dest Pointer to set of destination sites.
 * @param potential Pointer to array of potential in grid to write the result to.
 * @param max_iter Maximum number of iterations.
 * @param tolerance Tolerance of solver.
 * */
CGSolver::CGSolver(int width, int height, std::unordered_set<int> * grown, std::unordered_set<int> * dest, double * potential, int max_iter, double tolerance)
{
	this->width = width;
	this->height = height;
	this->adaptive = false;
	this->solverType = SOLVER_EIGEN_REGULAR;

	this->x = Eigen::VectorXd::Zero(this->width * this->height);
	this->x_old = Eigen::VectorXd::Zero(this->width * this->height);
	this->b = Eigen::VectorXd::Zero(this->width * this->height);

	this->A = Eigen::SparseMatrix<double>(this->width * this->height, this->width * this->height);
	this->A.reserve(Eigen::VectorXi::Constant(this->width*this->height,5));
	this->cg.setMaxIterations(max_iter);
	this->cg.setTolerance(tolerance);	

	this->grown = grown;
	this->dest = dest;
	this->potential = potential;
	this->max_iter = max_iter;
	this->tolerance = tolerance;
	this->makeAMatrixRegular();
}

/**
 * Destructor.
 * */
CGSolver::~CGSolver()
{
	if(this->solverType == SOLVER_CUDA)
	{
		delete this->cudaSolver;
	}
}

/**
 * Find adaptive grid cell neighbour(s) and insert/edit temporary matrix values.
 * @param qTN Quadtree node -- central cell we are working with -- finding neighbours
 * @param direction Search direction: NORTH, EAST, SOUTH or WEST -- 8, 6, 2 or 4 respectively.
 * @param AA Temporary map data structure to hold entries for edits of A matrix: successive calls to \c setAMatrix() and therefore
 * \c SetAMatrixNeighbour() may want to change already inserted entries in the map.
 * */
void CGSolver::SetAMatrixNeighbour(QuadTreeNode * qTN, int direction, std::map<int,double> * AA)
{	
	//find neighbours
	QuadTreeNode * neighbour = NULL;
	QuadTreeNode * neighbour1 = NULL;
	QuadTreeNode * neighbour2 = NULL;
	switch(direction)
	{
	case NORTH:
		neighbour = qTN->getNorthNeighbour();
		break;
	case EAST:
		neighbour = qTN->getEastNeighbour();
		break;
	case SOUTH:
		neighbour = qTN->getSouthNeighbour();
		break;
	case WEST:
		neighbour = qTN->getWestNeighbour();
		break;
	}
	if (neighbour == NULL)
	{
		//A.coeffRef(qTN->topLeft,qTN->topLeft) += 1.0f;
		std::map<int, double>::iterator it = AA->find(qTN->topLeft);
		switch(direction)
		{
		case NORTH:
		case SOUTH:
			if (it != AA->end())
			{
				it->second += 0.1;
			}
			else
			{
				AA->insert(std::pair<int, double>(qTN->topLeft,0.1));
			}
			break;
		case WEST:
		case EAST:
			if (it != AA->end())
			{
				it->second += 1.0;
			}
			else
			{
				AA->insert(std::pair<int, double>(qTN->topLeft,1.0));
			}
			break;
		}
		return;
		
	}
	if (neighbour->isLeaf())
	{
		//same level
		if (neighbour->level == qTN->level)
		{
			//if this is destination site, therefore BC with pot = 1
			std::map<int, double>::iterator it = AA->find(qTN->topLeft);
			if (it != AA->end())
			{
				it->second += 1.0;
			}
			else
			{
				AA->insert(std::pair<int, double>(qTN->topLeft,1.0));
			}
			if (this->dest->find(neighbour->topLeft) != this->dest->end())
			{
				b(qTN->topLeft) += 1.0f;
			}
			//growth site, therefore BC with pot = 0
			else if(this->grown->find(neighbour->topLeft) != this->grown->end())
			{
				
			}
			//ordinary site where potential is to be calculated
			else
			{
				AA->insert(std::pair<int, double>(neighbour->topLeft,-1.0));
			}			
			if(this->grown_last->find(neighbour->topLeft) != this->grown_last->end())
			{
				b(qTN->topLeft) -= DEPOSIT;
			}
		}
		//larger
		else
		{
			std::map<int, double>::iterator it = AA->find(qTN->topLeft);
			if (it != AA->end())
			{
				it->second += 0.5;
			}
			else
			{
				AA->insert(std::pair<int, double>(qTN->topLeft,0.5));
			}

			if (this->dest->find(neighbour->topLeft) != this->dest->end())
			{
				b(qTN->topLeft) += 0.5f;
			}
			//growth site, therefore BC with pot = 0
			else if(this->grown->find(neighbour->topLeft) != this->grown->end())
			{
						
			}
			//ordinary site where potential is to be calculated
			else
			{
				AA->insert(std::pair<int, double>(neighbour->topLeft,-0.5));
			}
			if(this->grown_last->find(neighbour->topLeft) != this->grown_last->end())
			{
				b(qTN->topLeft) -= 0.5*DEPOSIT;
			}
		}
	}
	//smaller
	else
	{
		switch(direction)
		{
		case NORTH:
			neighbour1 = neighbour->SW;
			neighbour2 = neighbour->SE;
			break;
		case EAST:
			neighbour1 = neighbour->NW;
			neighbour2 = neighbour->SW;
			break;
		case SOUTH:
			neighbour1 = neighbour->NW;
			neighbour2 = neighbour->NE;
			break;
		case WEST:
			neighbour1 = neighbour->NE;
			neighbour2 = neighbour->SE;
			break;
		}
		//N1
		//if this is destination site, therefore BC with pot = 1
		std::map<int, double>::iterator it = AA->find(qTN->topLeft);
			if (it != AA->end())
			{
				it->second += 0.5;
			}
			else
			{
				AA->insert(std::pair<int, double>(qTN->topLeft,0.5));
			}
		if (this->dest->find(neighbour1->topLeft) != this->dest->end())
		{
			b(qTN->topLeft) += 0.5f;
		}
		//growth site, therefore BC with pot = 0
		else if(this->grown->find(neighbour1->topLeft) != this->grown->end())
		{
						
		}
		//ordinary site where potential is to be calculated
		else
		{
			AA->insert(std::pair<int, double>(neighbour1->topLeft,-0.5));
		}
		if(this->grown_last->find(neighbour1->topLeft) != this->grown_last->end())
		{
			b(qTN->topLeft) -= 0.5*DEPOSIT;
		}
		//N2
		//if this is destination site, therefore BC with pot = 1
		it = AA->find(qTN->topLeft);
			if (it != AA->end())
			{
				it->second += 0.5;
			}
			else
			{
				AA->insert(std::pair<int, double>(qTN->topLeft,0.5));
			}
		if (this->dest->find(neighbour2->topLeft) != this->dest->end())
		{
			b(qTN->topLeft) += 0.5f;
		}
		//growth site, therefore BC with pot = 0
		else if(this->grown->find(neighbour2->topLeft) != this->grown->end())
		{
						
		}
		//ordinary site where potential is to be calculated
		else
		{
			AA->insert(std::pair<int, double>(neighbour2->topLeft,-0.5));
		}
		if(this->grown_last->find(neighbour2->topLeft) != this->grown_last->end())
		{
			b(qTN->topLeft) -= 0.5*DEPOSIT;
		}
	}
}

/**
* Remove row in tripletlist sparse matrix implementation.
* @param row Index of row to remove.
* */
void CGSolver::removeRow(int row)
{
	if(this->rowBeginings[row] != this->tripletList.end())
	{
		std::list<T>::iterator it = this->rowBeginings[row];
		while(it != this->tripletList.end() && it->row() == row)
		{
			++it;
		}
		this->tripletList.erase(this->rowBeginings[row],it);
		this->rowBeginings[row] = this->tripletList.end();
	}
}

/**
* Update row of the tripletlist A matrix implementation.
* @param qTN Quadtree node -- grid cell -- correstponding to row.
* */
void CGSolver::SetAMatrix(QuadTreeNode * qTN)
{		
	b(qTN->topLeft) = 0;
	std::map<int,double> * AA = new std::map<int,double>();
	//site is not BC
	if(this->grown->find(qTN->topLeft) == this->grown->end() &&
			this->dest->find(qTN->topLeft) == this->dest->end())
	{
		// now insert is ok, won't be for editing		
		this->SetAMatrixNeighbour(qTN,NORTH,AA);
		this->SetAMatrixNeighbour(qTN,EAST,AA);
		this->SetAMatrixNeighbour(qTN,SOUTH,AA);
		this->SetAMatrixNeighbour(qTN,WEST,AA);
	}
	int size1 = AA->size();

	bool firstpushback = true;
	for (std::map<int, double>::iterator iit = AA->begin(); iit != AA->end(); ++iit)
	{
		this->tripletList.push_back(T(qTN->topLeft,iit->first,iit->second));
		if (firstpushback)
		{
			std::list<T>::iterator tmpit = this->tripletList.end();
			--tmpit;
			this->rowBeginings[qTN->topLeft] = tmpit;
			firstpushback = false;
		}
	}
	delete AA;
	
}

/**
* Construct A matrix at the begining of the adaptive grid simulation.
*/
void CGSolver::makeAMatrix()
{
	for (std::set<QuadTreeNode *>::iterator it = this->quadtree->leaves->begin(); it != this->quadtree->leaves->end(); ++it)
	{
		if(*it != NULL)
		{
			this->removeRow((*it)->topLeft);
			this->SetAMatrix(*it);
		}
	}
	if(this->solverType == SOLVER_EIGEN)
	{
		this->A.setFromTriplets(this->tripletList.begin(), this->tripletList.end());
	}
	else if(this->solverType == SOLVER_CUDA)
	{
		this->makeAMatrixCuda();
	}
}

/**
* Construct A matrix at the begining of the regular grid simulation.
*/
void CGSolver::makeAMatrixRegular()
{
	for (int i=0; i<this->height; ++i)
	{
		for (int j=0; j<this->width; ++j)
		{
			//site is not BC
			if(this->grown->find(i*this->width + j) == this->grown->end() &&
				this->dest->find(i*this->width + j) == this->dest->end())
			{
				b(i*this->width+j) = 0;
				A.insert(i*this->width+j,i*this->width+j) = 1;
				//SOUTH
				//if(i<HEIGHT-1)
				if((i+1)*this->width+j < this->width*this->height)
				{
					//if this is destination site, therefore BC with pot = 1
					if(this->dest->find((i+1)*this->width + j) != this->dest->end())
					{
						b(i*this->width+j) += 0.25f;
					}
					//growth site, therefore BC with pot = 0
					else if(this->grown->find((i+1)*this->width + j) != this->grown->end())
					{
						
					}
					//ordinary site where potential is to be calculated
					else
					{
						A.insert(i*this->width+j,(i+1)*this->width+j) = -0.25f;
					}
				}
				//NORTH
				if(i>0)
				{
					//if this is destination site, therefore BC with pot = 1
					if(this->dest->find((i-1)*this->width + j) != this->dest->end())
					{
						b(i*this->width+j) += 0.25f;
					}
					//growth site, therefore BC with pot = 0
					else if(this->grown->find((i-1)*this->width + j) != this->grown->end())
					{
						
					}
					//ordinary site where potential is to be calculated
					else
					{
						A.insert(i*this->width+j,(i-1)*this->width+j) = -0.25f;
					}
				}
				//EAST
				//if(j<WIDTH-1)
				if(i*this->width+(j+1) < this->width * this->height)
				{
					//if this is destination site, therefore BC with pot = 1
					if(this->dest->find(i*this->width + (j+1)) != this->dest->end())
					{
						b(i*this->width+j) += 0.25f;
					}
					//growth site, therefore BC with pot = 0
					else if(this->grown->find(i*this->width + (j+1)) != this->grown->end())
					{
						
					}
					//ordinary site where potential is to be calculated
					else
					{
						A.insert(i*this->width+j,i*this->width+(j+1)) = -0.25f;
					}
				}
				//WEST
				if(j>0)
				{
					//if this is destination site, therefore BC with pot = 1
					if(this->dest->find(i*this->width + (j-1)) != this->dest->end())
					{
						b(i*this->width+j) += 0.25f;
					}
					//growth site, therefore BC with pot = 0
					else if(this->grown->find(i*this->width + (j-1)) != this->grown->end())
					{
						
					}
					//ordinary site where potential is to be calculated
					else
					{
						A.insert(i*this->width+j,i*this->width+(j-1)) = -0.25f;
					}
				}
			}
			else
			{
				b(i*this->width+j) = 0;
			}
		}
	}
}

/**
* Update A matrix between iterations of the regular grid simulation.
* @param nG Last growth site.
*/
void CGSolver::updateAMatrixRegular(int nG)
{
	//make row & column of new BC zero
	A.coeffRef(nG,nG) = 0;
	if(nG+1 < this->width * this->height)
	{
	  A.coeffRef(nG+1,nG) = 0;
	  A.coeffRef(nG,nG+1) = 0;
	}
	if(nG-1 >= 0)
	{
	  A.coeffRef(nG-1,nG) = 0;
	  A.coeffRef(nG,nG-1) = 0;
	}	
	//not sure about W/H in relation to i/j
	//assumes this->width = this->height
	if(nG + this->width < this->width * this->height)
	{
	  A.coeffRef(nG+this->width,nG) = 0;
	  A.coeffRef(nG,nG+this->width) = 0;
	}
	if(nG - this->width >=0 )
	{
	  A.coeffRef(nG-this->width,nG) = 0;
	  A.coeffRef(nG,nG-this->width) = 0;
 	}
	//edit b vector
	b(nG) = 0;
	if(nG+1 < b.size() && this->dest->find(nG+1) != this->dest->end())
	{
		b(nG+1) -= 0.25f;
	}
	if(nG > 0 && this->dest->find(nG-1) != this->dest->end())
	{
		b(nG-1) -= 0.25f;
	}
	if(nG+this->width < b.size() && this->dest->find(nG+this->width) != this->dest->end())
	{
		b(nG+this->width) -= 0.25f;
	}
	if(nG-this->width >= 0 && this->dest->find(nG+this->width) != this->dest->end())
	{
		b(nG-this->width) -= 0.25f;
	}	
}

/**
* Perform first iteration of solver
* @return 0 on success, -1 if number of iterations exceeded max_iter
* */
int CGSolver::solvePotential()
{
	int res = 0;
	if (this->solverType == SOLVER_EIGEN || this->solverType == SOLVER_EIGEN_REGULAR)
	{
		res = this->doSolvingEigen();
	}
	else if (this->solverType == SOLVER_PROG2)
	{
		res = this->solveProg2b();
	}
	else if (this->solverType == SOLVER_CUDA)
	{
		res = this->cudaSolver->cudaSolve();
		this->cudaSolver->cudaCopyResult();
		for(int i = 0; i< this->width * this->height; ++i)
		{
			if(this->dest->find(i) != this->dest->end())
			{
				this->potential[i] = 1.0;
			}
			else
			{
				this->potential[i] = this->cudaSolver->h_x[i];
			}
		}
		this->cudaSolver->cudaDestroyAMatrixCuda();
	}
	if(this->solverType != SOLVER_CUDA)
	{
		this->reWriteResult();
	}
	return res;
}

/**
* Perform subsequent iteration of solver. Prior to the solving itself, the A matrix is updated based on the new growth site and list of subdivided cells in quadtree.
* @param newGrown Last growth site.
* @return 0 on success, -1 if number of iterations exceeded max_iter.
* */
int CGSolver::solvePotential(int newGrown)
{
	int res = 0;
	if(this->adaptive)
	{
		//this->printAMatrix("old");
		//add new grown to list
		//neighbours have to be same size or bigger
		//and they have to be leaves because this node is of the highest level
		QuadTreeNode * newGrownNode = this->quadtree->find(newGrown);
		this->quadtree->subdivided->emplace(newGrownNode);
	
		this->quadtree->subdivided->emplace(newGrownNode->getNorthNeighbour());
		this->quadtree->subdivided->emplace(newGrownNode->getEastNeighbour());
		this->quadtree->subdivided->emplace(newGrownNode->getSouthNeighbour());
		this->quadtree->subdivided->emplace(newGrownNode->getWestNeighbour());

		for(std::set<QuadTreeNode*>::iterator it = this->quadtree->subdivided->begin(); it != this->quadtree->subdivided->end(); ++it)
		{
			if((*it) != NULL)
			{			
				if((*it)->isLeaf())
				{
					this->removeRow((*it)->topLeft);
					this->SetAMatrix(*it);				
				}		
			}				
		}
		this->quadtree->subdivided->clear();
	
		if (this->solverType == SOLVER_EIGEN)
		{
			this->A.setFromTriplets(this->tripletList.begin(), this->tripletList.end());
		}
	}
	else
	{
		this->updateAMatrixRegular(newGrown);
	}

	if (this->solverType == SOLVER_EIGEN || this->solverType == SOLVER_EIGEN_REGULAR)
	{
		res = this->doSolvingEigen();
		this->reWriteResult();
	}
	else if (this->solverType == SOLVER_PROG2)
	{
		res = this->solveProg2b();
		this->reWriteResult();	
	}
	else if (this->solverType == SOLVER_CUDA)
	{
		this->makeAMatrixCuda();
		res = this->cudaSolver->cudaSolve();
		this->cudaSolver->cudaCopyResult();
		for(int i = 0; i< this->width * this->height; ++i)
		{
			if(this->dest->find(i) != this->dest->end())
			{
				this->potential[i] = 1.0;
			}
			else
			{
				this->potential[i] = this->cudaSolver->h_x[i];
			}
		}
		this->cudaSolver->cudaDestroyAMatrixCuda();
	}	
	return res;
}

/**
* Run the Eigen CG solver.
* @return 0 on success, -1 if number of iterations exceeded max_iter or decomposition fails.
* */
int CGSolver::doSolvingEigen()
{
	double local_tolerance;
	if(frameno == 0)
	{
		local_tolerance = tolerance;
	}
	else
	{
		local_tolerance = 1000 * tolerance;
	}
	this->cg.setTolerance(local_tolerance);

	this->cg.compute(this->A);
	if(this->cg.info() != Eigen::Success)
	{
		std::cout << "decomposition failed\n";
		return -1;
	}
	this->x_old = this->x;
	this->x = this->cg.solveWithGuess(this->b,this->x_old);
	//this->x = this->cg.solve(this->b);
	if(this->cg.info() != Eigen::Success)
	{
		std::cout << "solving failed\n";
		std::cout << cg.info() << "\n";
		
		return -1;
	}
	else
	{
		//std::cout << "#iterations: " << cg.iterations() << std::endl;
		//std::cout << "error: " << cg.error() << std::endl;
		return 0;
	}
}

/**
 * Write resulting \c x vector to \c potential array.
 * */
void CGSolver::reWriteResult()
{
	for(int i = 0; i<x.size(); ++i)
	{
		if(this->dest->find(i) != this->dest->end())
		{
			this->potential[i] = 1.0;
		}
		else
		{
			this->potential[i] = x(i);
		}		
	}
}

/**
 * Sparse matrix-vector multiplication y = Ax.
 * Matrix is stored as Eigen triplet list with a vector of row beginnings.
 * Uses OpenMP for parallelization.
 * @param x Vector x
 * @param y Resulting vector y 
 * */
void CGSolver::Ax(Eigen::VectorXd &x, Eigen::VectorXd &y)
{
#pragma omp parallel for
	for (int i = 0; i < this->width * this->height; ++i)
	{
		y.coeffRef(i) = 0;
		if (this->rowBeginings[i] != this->tripletList.end())
		{
			std::list<T>::iterator it = this->rowBeginings[i];
			while (it->row() == i)
			{				
				y.coeffRef(i) += it->value() * x.coeffRef(it->col());
				++it;
			}
		}
	}
}

/**
 * Run own CG solver.
 * Own basic implementation of Conjugate Gradient Solver based on Shewchuck's paper "An Introduction to the Conjugate Gradient Method Without the Agonizing Pain", page 32.
 * No real optimisations, no preconditioning. Vector operations and matrix-vector multiplication parallelised
 * @return 0 on success, -1 if number of iterations exceeded max_iter.
 * */
int CGSolver::solveProg2b()
{
	double alpha = 0;
	double beta = 0;
	int k = 0;
	this->Ax(this->x,this->d);
	this->r = this->b - this->d;
	this->d = this->r;

	double local_tolerance;

	if(frameno == 0)
	{
		//local_tolerance = 0.001 * tolerance;
		local_tolerance = tolerance;
	}
	else
	{
		//local_tolerance = 10 * tolerance * tolerance;
		local_tolerance = 1000 * tolerance;
	}

	while (this->r.squaredNorm() >local_tolerance && k < this->max_iter)
	{
		this->Ax(this->d,this->Ad);
		alpha = this->r.dot(this->r) / this->d.dot(this->Ad);
		this->x += alpha * this->d;
		this->r_old = this->r;
		this->r -= alpha * Ad;
		beta = this->r.dot(this->r) / this->r_old.dot(this->r_old);
		this->d = this->r + beta * this->d;
		++k;
	}
	//std::cout << "#iterations: " << k << std::endl;
	//std::cout << "error: " << this->r.squaredNorm() << std::endl;
	if(k >= this->max_iter)
	{
		return -1;
	}
	else
	{
		return 0;
	}
}

/**
 * Prepare A matrix in CSR format for CUDA, created from triplet list.
 * */
void CGSolver::makeAMatrixCuda()
{
	this->I = (int *)malloc(sizeof(int) * (this->width * this->height + 1));
	this->J = (int *)malloc(sizeof(int) * this->tripletList.size());
	this->val = (double *)malloc(sizeof(double) * this->tripletList.size());
	
	int indexcounter = 0;
	int colscounter = 0;
	int valscounter = 0;
	for(int i = 0; i < this->width * this->height; ++i)
	{
		I[i] = indexcounter;
		if(this->rowBeginings[i] != this->tripletList.end())
		{
			colscounter = 0;
			std::list<T>::iterator it = this->rowBeginings[i];
			while(it != this->tripletList.end() && it->row() == i)
			{
				val[valscounter] = it->value();
				J[valscounter] = it->col();
				++indexcounter;
				++colscounter;
				++valscounter;
				++it;
			}			
		}		
	}
	this->I[this->width*this->height]=this->tripletList.size() +  this->I[0];
	for(int i = 0; i< this->width * this->height; ++i)
	{
		this->cudaSolver->h_b[i] = this->b.coeff(i);
	}
	this->cudaSolver->cudaMakeAMatrixCuda(this->I,this->J,this->val);	
}