/*
 * P E R U N -- a physically based lightning generator
 * Master's thesis
 * Department of Computer Graphics and Interaction
 * Faculty of Electrical Engineering
 * Czech Technical University in Prague.
 *
 * Supervisor: Ing. Jaroslav Sloup
 *
 * Distributed under the GNU General Public License v3.0
 * */

/**
 * @file
 * @author Richard Dobrichovsky <dobrichovsky@gmail.com>
 * @date May 2016
 * 
 * @brief Declarations for the CGSolver class.
 *
 * Declarations for the CGSolver class which includes 4 different solver options.
 *
 * */

#include <time.h>
#include <unordered_set>
#include <iostream>
#include <fstream>

#include "common.h"

#include <Eigen/Sparse>
#include <Eigen/Dense>
#include <Eigen/IterativeLinearSolvers>
#include <Eigen/Core>

#ifndef cg_solver_cuda_inc
#define cg_solver_cuda_inc
#include "cg_solver_cuda.cuh"
#endif

typedef Eigen::Triplet<double> T;

#include "quadtree.h"

/**
 * 
 * Class which implements solvers utilising the method of conjugate gradients.
 * @brief Class which implements solvers utilising the method of conjugate gradients.
 * */
class CGSolver 
{
public:
	CGSolver(int width, int height, QuadTree * quadtree, std::unordered_set<int> * grown, std::unordered_set<int> * grown_last, std::unordered_set<int> * dest, double * potential, int max_iter, double tolerance, int solverType);
	CGSolver(int width, int height, std::unordered_set<int> * grown, std::unordered_set<int> * dest, double * potential, int max_iter, double tolerance);
	~CGSolver();
	int solvePotential();
	int solvePotential(int newGrown);
	void makeAMatrix();
	void removeRow(int row);

	void makeAMatrixRegular();
	void updateAMatrixRegular(int newGrown);

	///List of Eigen triplets: an editing-effective way of representing a sparse matrix.
	std::list<T> tripletList;
	///Pointers to sparse matrix row beginings.
	std::vector<std::list<T>::iterator> rowBeginings;	
	///Maximum number of iterations.
	int max_iter;
	///Tolerance of solver.
	double tolerance;

	/* CUDA stuff */
	void makeAMatrixCuda();
	///Array of row indices for CSR sparse matrix format.
	int * I;
	///Array of column indices for CSR sparse matrix format.
	int * J;
	///Array of values for CSR sparse matrix format.
	double * val;
	///Pointer to CUDA solver object.
	CudaSolver * cudaSolver;
	/* End CUDA stuff */

private:
	void reWriteResult();

	///Pointer to growth sites set.
	std::unordered_set<int> * grown;
	///Pointer to destination sites set.
	std::unordered_set<int> * dest;
	///Pointer to growth sites set in previous frame.
	std::unordered_set<int> * grown_last;
	///Calculated potential in simulation.
	double * potential;

	///Solver type: SOLVER_EIGEN_REGULAR, SOLVER_EIGEN, SOLVER_PROG2 or SOLVER_CUDA -- 0, 1, 2 or 3 respectively.
	int solverType;
	///Whether we are using adaptive grid.
	bool adaptive;

	///Vector of simulation results.
	Eigen::VectorXd x;
	///Right-side vector Ax=b.
	Eigen::VectorXd	b;
	///Vector of simulation results in previous iteration.
	Eigen::VectorXd x_old;

	/* Prog2 stuff */
	///Prog2 residual vector.
	Eigen::VectorXd r;
	///Prog2 residual vector in previous iteration.
	Eigen::VectorXd r_old;
	///Prog2 d vector in CG method.
	Eigen::VectorXd d;
	///Prog2 Matrix-vector multiplication Ad result.
	Eigen::VectorXd Ad;

	void Ax(Eigen::VectorXd &x, Eigen::VectorXd &y);
	int solveProg2b();
	/* End Prog2 stuff */

	/* Eigen stuff */
	int doSolvingEigen();
	///Eigen sparse matrix A.
	Eigen::SparseMatrix<double> A;
	///Eigen conjugate gradient method handler.
	Eigen::ConjugateGradient<Eigen::SparseMatrix<double>,Eigen::Lower|Eigen::Upper> cg;
	/* End Eigen stuff */	

	/* Adaptive stuff */
	///Quadtree object.
	QuadTree * quadtree;
	void SetAMatrix(QuadTreeNode * quadTreeNode);
	void SetAMatrixNeighbour(QuadTreeNode * quadTreeNode, int direction, std::map<int,double> * AA);
	/* End adaptive stuff */

	///Simulation grid width
	int width;
	///Simulation grid height
	int height;
};