/*
 * P E R U N -- a physically based lightning generator
 * Master's thesis
 * Department of Computer Graphics and Interaction
 * Faculty of Electrical Engineering
 * Czech Technical University in Prague.
 *
 * Supervisor: Ing. Jaroslav Sloup
 *
 * Distributed under the GNU General Public License v3.0
 * */

/**
 * @file
 * @author Richard Dobrichovsky <dobrichovsky@gmail.com>
 * @date May 2016
 * 
 * @brief Definitions for the CudaSolver class.
 *
 * An implementation of a conjugate gradient solver in CUDA. This implementation uses the cuBlas library for linear algebra kernels
 * and cuSparse for representing the A matrix in CSR format and subsequent matrix-vector multiplications.
 *
 * */

#include "cg_solver_cuda.cuh"
#include <stdio.h>
#include <stdlib.h>

#include <vector>
#include <list>

// Simple function to check for CUDA runtime errors.
static void handleCUDAError(
	cudaError_t error,		// error code
	const char *file,		// file within error was generated
	int line )			// line where error occurs
{
  if (error != cudaSuccess) {	// any error -> display error message and terminate application
    printf( "%s in %s at line %d\n", cudaGetErrorString( error ), file, line );
    exit( EXIT_FAILURE );
  }
}

#define CHECK_ERROR( error ) ( handleCUDAError( error, __FILE__, __LINE__ ) )

static const char *_cudaGetErrorEnum(cusparseStatus_t error)
{
    switch (error)
    {
        case CUSPARSE_STATUS_SUCCESS:
            return "CUSPARSE_STATUS_SUCCESS";

        case CUSPARSE_STATUS_NOT_INITIALIZED:
            return "CUSPARSE_STATUS_NOT_INITIALIZED";

        case CUSPARSE_STATUS_ALLOC_FAILED:
            return "CUSPARSE_STATUS_ALLOC_FAILED";

        case CUSPARSE_STATUS_INVALID_VALUE:
            return "CUSPARSE_STATUS_INVALID_VALUE";

        case CUSPARSE_STATUS_ARCH_MISMATCH:
            return "CUSPARSE_STATUS_ARCH_MISMATCH";

        case CUSPARSE_STATUS_MAPPING_ERROR:
            return "CUSPARSE_STATUS_MAPPING_ERROR";

        case CUSPARSE_STATUS_EXECUTION_FAILED:
            return "CUSPARSE_STATUS_EXECUTION_FAILED";

        case CUSPARSE_STATUS_INTERNAL_ERROR:
            return "CUSPARSE_STATUS_INTERNAL_ERROR";

        case CUSPARSE_STATUS_MATRIX_TYPE_NOT_SUPPORTED:
            return "CUSPARSE_STATUS_MATRIX_TYPE_NOT_SUPPORTED";
    }

    return "<unknown>";
}

static const char *_cudaGetErrorEnum(cublasStatus_t error)
{
    switch (error)
    {
        case CUBLAS_STATUS_SUCCESS:
            return "CUBLAS_STATUS_SUCCESS";

        case CUBLAS_STATUS_NOT_INITIALIZED:
            return "CUBLAS_STATUS_NOT_INITIALIZED";

        case CUBLAS_STATUS_ALLOC_FAILED:
            return "CUBLAS_STATUS_ALLOC_FAILED";

        case CUBLAS_STATUS_INVALID_VALUE:
            return "CUBLAS_STATUS_INVALID_VALUE";

        case CUBLAS_STATUS_ARCH_MISMATCH:
            return "CUBLAS_STATUS_ARCH_MISMATCH";

        case CUBLAS_STATUS_MAPPING_ERROR:
            return "CUBLAS_STATUS_MAPPING_ERROR";

        case CUBLAS_STATUS_EXECUTION_FAILED:
            return "CUBLAS_STATUS_EXECUTION_FAILED";

        case CUBLAS_STATUS_INTERNAL_ERROR:
            return "CUBLAS_STATUS_INTERNAL_ERROR";
    }

    return "<unknown>";
}

template< typename T >
void check(T result, char const *const func, const char *const file, int const line)
{
    if (result)
    {
        fprintf(stderr, "CUDA error at %s:%d code=%d(%s) \"%s\" \n",
                file, line, static_cast<unsigned int>(result), _cudaGetErrorEnum(result), func);
        //DEVICE_RESET
		cudaDeviceReset();
        // Make sure we call CUDA Device Reset before exiting
        exit(EXIT_FAILURE);
    }
}

#define CHECK_ERROR_2(val)           check ( (val), #val, __FILE__, __LINE__ )

/**
 * Constructor. Calls \c initCuSparse() to allocate resourses.
 * @param tolerance Solver tolerance.
 * @param iterations Maximum number of solver iterations.
 * @param width Simulation grid width.
 * @param height Simulation grid height.
 * */
CudaSolver::CudaSolver(double tolerance, int iterations, int width, int height)
{
	this->width = width;
	this->height = height;
	this->tolerance = tolerance;
	this->iterations = iterations;
	this->initCuSparse();
}

/**
 * Destructor. Frees all allocated resources, both host and device.
 * */
CudaSolver::~CudaSolver()
{
	free(this->h_x);
	free(this->h_b);
	cudaFree(this->d_x);
	cudaFree(this->d_b);
	cudaFree(this->d_r);
	cudaFree(this->d_r_old);
	cudaFree(this->d_p);
	cudaFree(this->d_Ax);
	cudaFree(this->d_Ad);
	cudaDeviceReset();
}

/**
 * Allocate necessary resources, both on host and device side.
 * */
void CudaSolver::initCuSparse()
{
	/* Get handle to the CUBLAS context */
    this->cublasHandle = 0;
    this->cublasStatus = cublasCreate(&cublasHandle);
	CHECK_ERROR_2(this->cublasStatus);

    /* Get handle to the CUSPARSE context */
    this->cusparseHandle = 0;
    this->cusparseStatus = cusparseCreate(&cusparseHandle);
	CHECK_ERROR_2(this->cusparseStatus);

    this->descr = 0;
    this->cusparseStatus = cusparseCreateMatDescr(&descr);
	CHECK_ERROR_2(this->cusparseStatus);

    this->cusparseStatus = cusparseSetMatType(this->descr,CUSPARSE_MATRIX_TYPE_GENERAL);
    this->cusparseStatus = cusparseSetMatIndexBase(this->descr,CUSPARSE_INDEX_BASE_ZERO);
	CHECK_ERROR_2(this->cusparseStatus);

	cublasSetPointerMode(cublasHandle,CUBLAS_POINTER_MODE_DEVICE);
	cusparseSetPointerMode(cusparseHandle,CUSPARSE_POINTER_MODE_DEVICE);

	CHECK_ERROR( cudaGetLastError() );
	this->N = this->width * this->height;
	
	this->h_x = (double *)malloc(sizeof(double)*N);
    this->h_b = (double *)malloc(sizeof(double)*N);

	for (int i = 0; i < N; i++)
    {
        h_b[i] = 0.0;
        h_x[i] = 0.0;
    }

	CHECK_ERROR(cudaMalloc((void **)&d_x, N*sizeof(double)));
    CHECK_ERROR(cudaMalloc((void **)&d_b, N*sizeof(double)));
	CHECK_ERROR(cudaMalloc((void **)&d_r, N*sizeof(double)));
	CHECK_ERROR(cudaMalloc((void **)&d_r_old, N*sizeof(double)));
    CHECK_ERROR(cudaMalloc((void **)&d_p, N*sizeof(double)));
    CHECK_ERROR(cudaMalloc((void **)&d_Ax, N*sizeof(double)));
	CHECK_ERROR(cudaMalloc((void **)&d_Ad, N*sizeof(double)));

	CHECK_ERROR(cudaMalloc((void **)&alpha, sizeof(double)));	
	CHECK_ERROR(cudaMalloc((void **)&alpha1, sizeof(double)));
	CHECK_ERROR(cudaMalloc((void **)&beta, sizeof(double)));	
	CHECK_ERROR(cudaMalloc((void **)&beta1, sizeof(double)));

	CHECK_ERROR(cudaMalloc((void **)&d_alpha, sizeof(double)));
	CHECK_ERROR(cudaMalloc((void **)&d_alpha_m, sizeof(double)));
	CHECK_ERROR(cudaMalloc((void **)&d_beta, sizeof(double)));

	double aaa = 1.0;
	CHECK_ERROR(cudaMemcpy(d_alpha, &aaa, sizeof(double), cudaMemcpyHostToDevice));
	aaa = -1.0;
	CHECK_ERROR(cudaMemcpy(d_alpha_m, &aaa, sizeof(double), cudaMemcpyHostToDevice));
	aaa = 0.0;
	CHECK_ERROR(cudaMemcpy(d_beta, &aaa, sizeof(double), cudaMemcpyHostToDevice));
}

/**
 * Allocate space and copy matrix in CSR format to device.
 * @param II Host array of row indices for CSR sparse matrix format.
 * @param JJ Host array of column indices for CSR sparse matrix format.
 * @param vval Host array of values for CSR sparse matrix format.
 */
void CudaSolver::cudaMakeAMatrixCuda(int * II, int * JJ, double * vval)
{
	this->I = II;
	this->J = JJ;
	this->val = vval;
	this->nnz = I[N] - I[0];

	CHECK_ERROR(cudaMalloc((void **)&d_col, nnz * sizeof(int)));
	CHECK_ERROR(cudaMalloc((void **)&d_row, (N+1)*sizeof(int)));
    CHECK_ERROR(cudaMalloc((void **)&d_val, nnz*sizeof(double)));

	CHECK_ERROR(cudaMemcpyAsync(d_col, J, nnz*sizeof(int), cudaMemcpyHostToDevice));
    CHECK_ERROR(cudaMemcpyAsync(d_row, I, (N+1)*sizeof(int), cudaMemcpyHostToDevice));
    CHECK_ERROR(cudaMemcpyAsync(d_val, val, nnz*sizeof(double), cudaMemcpyHostToDevice));
    CHECK_ERROR(cudaMemcpyAsync(d_x, this->h_x, N*sizeof(double), cudaMemcpyHostToDevice));
	
    CHECK_ERROR(cudaMemcpyAsync(d_b, this->h_b, N*sizeof(double), cudaMemcpyHostToDevice));
	cudaDeviceSynchronize();
}

/**
 * Destroy resources needed for A matrix on host and device.
 * */
void CudaSolver::cudaDestroyAMatrixCuda()
{
	CHECK_ERROR(cudaGetLastError());
	cudaFree(this->d_col);
	cudaFree(this->d_row);
	cudaFree(this->d_val);
	free(this->I);
	free(this->J);
	free(this->val);
	CHECK_ERROR(cudaGetLastError());
}

/**
 * Simple division on host \c alpha = \c alpha / \c alpha1
 * @param alpha Dividend and quotient: \c alpha = \c alpha / \c alpha1
 * @param alpha1 Divisor
 * */
__global__ void cudaDiv(double * alpha, double * alpha1)
{
	int threadID = threadIdx.x + blockDim.x * blockIdx.x;
	
	if(threadID == 0)
	{
		alpha[0] = alpha[0] / alpha1[0];
	}
}

/**
 * Simple multiplication by -1
 * @param alpha \c alpha = \c -alpha
 * */
__global__ void cudaMinOneTimes(double * alpha)
{
	int threadID = threadIdx.x + blockDim.x * blockIdx.x;
	
	if(threadID == 0)
	{
		alpha[0] = -alpha[0];
	}
}

/**
 * CUDA CG solver.
 * @return 0 on success, -1 if maximum iterations were exceeded.
 * */
int CudaSolver::cudaSolve()
{
	dim3 gridSize(1, 1, 1);
	dim3 blockSize(THREADS_PER_BLOCK, 1, 1);
	
	int k = 0;
	double r_sqnorm = 1;
	double local_tolerance;

	if(frameno == 0)
	{
		local_tolerance = tolerance;
	}
	else
	{
		local_tolerance = 1000 * tolerance;
	}

	//my Ax is d
	// d = Ax
	this->cusparseStatus = cusparseDcsrmv(this->cusparseHandle,CUSPARSE_OPERATION_NON_TRANSPOSE, N, N, nnz, d_alpha, descr, d_val, d_row, d_col, d_x, d_beta, d_Ax);
	CHECK_ERROR_2(this->cusparseStatus);	
	// r = b
	CHECK_ERROR_2(cublasDcopy(this->cublasHandle, N, this->d_b, 1, this->d_r, 1));
	// r = b - d (r -= d)
	CHECK_ERROR_2(cublasDaxpy(this->cublasHandle, N, d_alpha_m, this->d_Ax, 1, this->d_r, 1));
	// d = r
	CHECK_ERROR_2(cublasDcopy(this->cublasHandle, N, this->d_r, 1, this->d_Ax, 1));
	//while (r_sqnorm > 10 * tolerance && k < iterations)
	while (r_sqnorm * r_sqnorm > local_tolerance && k < this->iterations)
	{		
		// Ad = A*Ax		
		CHECK_ERROR_2(cusparseDcsrmv(this->cusparseHandle,CUSPARSE_OPERATION_NON_TRANSPOSE, N, N, nnz, d_alpha, descr, d_val, d_row, d_col, d_Ax, d_beta, d_Ad));		
		CHECK_ERROR_2(cublasDdot(this->cublasHandle, N, this->d_r, 1, this->d_r, 1, alpha));
		CHECK_ERROR_2(cublasDdot(this->cublasHandle, N, this->d_Ax, 1, this->d_Ad, 1, alpha1));
		//alpha /= alpha1;
		cudaDiv<<<gridSize, blockSize>>>(alpha,alpha1);
		CHECK_ERROR_2(cublasDaxpy(this->cublasHandle, N, alpha, this->d_Ax, 1, this->d_x, 1));
		CHECK_ERROR_2(cublasDcopy(this->cublasHandle, N, this->d_r, 1, this->d_r_old, 1));
		//alpha *= -1.0f;
		cudaMinOneTimes<<<gridSize, blockSize>>>(alpha);
		CHECK_ERROR_2(cublasDaxpy(this->cublasHandle, N, alpha, this->d_Ad, 1, this->d_r, 1));
		CHECK_ERROR_2(cublasDdot(this->cublasHandle, N, this->d_r, 1, this->d_r, 1, beta));
		CHECK_ERROR_2(cublasDdot(this->cublasHandle, N, this->d_r_old, 1, this->d_r_old, 1, beta1));
		cudaDiv<<<gridSize, blockSize>>>(beta,beta1);
		CHECK_ERROR_2(cublasDscal(this->cublasHandle, N, beta, this->d_Ax, 1));
		CHECK_ERROR_2(cublasDaxpy(this->cublasHandle, N, d_alpha, this->d_r, 1, this->d_Ax, 1));
		cublasSetPointerMode(cublasHandle,CUBLAS_POINTER_MODE_HOST);
		CHECK_ERROR_2(cublasDnrm2(this->cublasHandle, N, this->d_r, 1, &r_sqnorm));	
		cublasSetPointerMode(cublasHandle,CUBLAS_POINTER_MODE_DEVICE);
		++k;
	}

	//printf("%i iterations\n",k);
	//printf("#iterations: %i\n",k);
	//printf("error: %f\n",r_sqnorm* r_sqnorm );
	

	CHECK_ERROR(cudaGetLastError());
	if(k >= 1000)
	{
		return -1;
	}
	else
	{
		return 0;
	}
}

/**
 * Copy results of solver from device to host: d_x to h_x
 * */
void CudaSolver::cudaCopyResult()
{
	cudaMemcpy(h_x, d_x, N*sizeof(double), cudaMemcpyDeviceToHost);
	CHECK_ERROR(cudaGetLastError());	
}