/*
 * P E R U N -- a physically based lightning generator
 * Master's thesis
 * Department of Computer Graphics and Interaction
 * Faculty of Electrical Engineering
 * Czech Technical University in Prague.
 *
 * Supervisor: Ing. Jaroslav Sloup
 *
 * Distributed under the GNU General Public License v3.0
 * */

/**
 * @file
 * @author Richard Dobrichovsky <dobrichovsky@gmail.com>
 * @date May 2016
 * 
 * @brief Declarations for the CudaSolver class.
 *
 * Declarations for the CudaSolver class which implements a conjugate gradient solver in CUDA.
 *
 * */

#include <cuda_runtime.h>
#include <cusparse.h>
#include <cublas_v2.h>

#include "common.h"

/// Threads per block
#define THREADS_PER_BLOCK 512

/**
 * CUDA implemetation of the method of conjugate gradients.
 * @brief CUDA implemetation of the method of conjugate gradients.
 * */
class CudaSolver
{
public:
	CudaSolver(double tolerance, int iterations, int width, int height);
	~CudaSolver();
	void initCuSparse();
	void cudaMakeAMatrixCuda(int * I, int * J, double * val);
	int cudaSolve();
	void cudaDestroyAMatrixCuda();
	void cudaCopyResult();

	///Simulation grid height
	int height;
	///Simulation grid width
	int width;
	///N = width * height
	int N;

	///Array of row indices for CSR sparse matrix format.
	int * I;
	///Array of column indices for CSR sparse matrix format.
	int * J;
	///Array of values for CSR sparse matrix format.
	double * val;

	///Host b vector Ax=b
	double * h_b;
	///Host x vector of results Ax=b
	double * h_x;
	
	///Device x vector of results Ax=b
	double * d_x;
	
	///Total number of non-zeros in A matrix
	int nnz;
		
private:
	///cuBlas handle
	cublasHandle_t cublasHandle;
	///cuBlas status
    cublasStatus_t cublasStatus;
	///cuSparse handle
	cusparseHandle_t cusparseHandle;
	///cuSparse status
    cusparseStatus_t cusparseStatus;
	///cuSparse matrix descriptor
    cusparseMatDescr_t descr;

	///Alpha in CG method
	double * alpha;
	///Alpha denominator in CG method
	double * alpha1;
	///Beta in CG method
	double * beta;
	///Beta denominator in CG method
	double * beta1;

	///Device variable d_alpha = 1.0
	double * d_alpha;
	///Device variable d_beta = 0.0
	double * d_beta;
	///Device variable d_alpha_m = -1.0
	double * d_alpha_m;
	
	///Device array of column indices for CSR sparse matrix format.
	int * d_col;
	///Device array of row indices for CSR sparse matrix format.
	int * d_row;
	///Device array of values for CSR sparse matrix format.
	double * d_val;
	///Device r vector of residuals.
	double * d_r;
	///Device r vector of residuals in previous iteration.
	double * d_r_old;
	///Device right side b vector Ax=b.
	double * d_b;
	///Device d vector in CG method.
	double * d_p;
	///Device Matrix-vector multiplication Ax result.
	double * d_Ax;
	///Device Matrix-vector multiplication Ad result.
	double * d_Ad;
	///Solver tolerance
	double tolerance;
	///Solver maximum number of iterations
	int iterations;

};


