/*
 * P E R U N -- a physically based lightning generator
 * Master's thesis
 * Department of Computer Graphics and Interaction
 * Faculty of Electrical Engineering
 * Czech Technical University in Prague.
 *
 * Supervisor: Ing. Jaroslav Sloup
 *
 * Distributed under the GNU General Public License v3.0
 * */


/**
 * @file
 * @author Richard Dobrichovsky <dobrichovsky@gmail.com>
 * @date May 2016
 * 
 * @brief Definitions for the DAG class.
 *
 * Definitions for the DAG class implementing a graph data structure to represent the generated lightning shape.
 *
 * */

#include "graph.h"
#include "common.h"

std::default_random_engine generator;
std::uniform_real_distribution<double> distribution(0,1);
std::ofstream output;

/**
* Constructor for creating DAG nodes during simulation.
* @param id Node id referring to simulation grid pixel.
* @param parent This node's parent.
* @param y Node's y coordinate in simulation grid.
* @param x Node's x coordinate in simulation grid.
* @param width Simulation grid width.
* @param height Simulation grid height.
* @param resolutionMultiple Resolution multiplicator.
* */
DAG::DAG(int id, DAG * parent, int y, int x, int width, int height, int resolutionMultiple)
{	
	this->parent = parent;
	if(parent != NULL)
	{
		parent->children.push_back(this);
	}
	double jitterx = distribution(generator);
	double jittery = distribution(generator);
	if((x - 0.5 + jitterx) * resolutionMultiple > 0)
	{
		if((x - 0.5 + jitterx) * resolutionMultiple < width * resolutionMultiple - 1)
		{
			this->x = (x - 0.5 + jitterx) * resolutionMultiple;
		}
		else
		{
			this->x = width * resolutionMultiple - 1;
		}
	}
	else
	{
		this->x = 0;
	}
	if((y - 0.5 + jittery) * resolutionMultiple > 0)
	{
		if((y - 0.5 + jittery) * resolutionMultiple < height * resolutionMultiple - 1)
		{
			this->y = (y - 0.5 + jittery) * resolutionMultiple;
		}
		else
		{
			this->y = height * resolutionMultiple - 1;
		}
	}
	else
	{
		this->y = 0;
	}
	this->type = 0;
	this->luminance = 0.0f;
	this->id = id;
}

/**
* Constructor for creating DAG nodes when loading DAG from file.
* @param id Node id referring to simulation grid pixel.
* @param x Node's x coordinate with jitter in image resolution.
* @param y Node's y coordinate with jitter in image resolution.
* @param type Node type (1, 2 or 3 -- main channel, secondary or ternary branch respectively).
* @param luminance Node luminance.
* */
DAG::DAG(int id, double x, double y, int type, double luminance)
{
	this->id = id;
	this->x = x;
	this->y = y;
	this->type = type;
	this->luminance = luminance;
}

/**
* Destructor. Deletes all children recursively.
* */
DAG::~DAG()
{
	for(std::vector<DAG*>::iterator it = this->children.begin(); it != this->children.end(); ++it)
	{
		delete *it;
	}
}

/**
* Save graph to file.
* @param frame Frame number.
* */
void DAG::saveGraph(int frame)
{
	std::string fname = "lightning_graph_";
	if (frame < 10)
	{
		fname += "000";		
	}
	else if (frame < 100)
	{
		fname += "00";
	}
	else if (frame < 1000)
	{
		fname += "0";
	}
	std::string fnumber = std::to_string(static_cast<long long>(frame));
	fname += fnumber;
	output.open(fname, std::ofstream::out | std::ofstream::trunc);
	output << "vertices\n";
	this->saveVertex();
	output << -1 << " edges\n";
	this->saveEdge();
	output << -1;
	output.close();
	//std::cout << "File saved: " << filename << "\n";
}

/**
* Write vertex to file.
* */
void DAG::saveVertex()
{
	output << this->id << " " << this->x << " " << this->y << " " << this->type << " " << this->luminance << "\n";
	for (std::vector<DAG*>::iterator it = this->children.begin(); it != this->children.end(); ++it)
	{
		(*it)->saveVertex();
	}
}

/**
* Write edge to file.
* */
void DAG::saveEdge()
{
	for(std::vector<DAG*>::iterator it = this->children.begin(); it != this->children.end(); ++it)
	{
		output << this->id << " " << (*it)->id << "\n";
		(*it)->saveEdge();
	}
}