/*
 * P E R U N -- a physically based lightning generator
 * Master's thesis
 * Department of Computer Graphics and Interaction
 * Faculty of Electrical Engineering
 * Czech Technical University in Prague.
 *
 * Supervisor: Ing. Jaroslav Sloup
 *
 * Distributed under the GNU General Public License v3.0
 * */

/**
 * @file
 * @author Richard Dobrichovsky <dobrichovsky@gmail.com>
 * @date May 2016
 * @brief Definitions for ImageFP class.
 * Definitions for the ImageFP class -- for manipulating floating point EXR images. Uses TinyEXR header-only library.
 * */

#include "imageFP.h"
#include <iostream>

#ifndef TINYEXR_IMPLEMENTATION
	#define TINYEXR_IMPLEMENTATION
	#include "tinyexr.h"
#endif

int frameno = 0;

/**
 * Save FP data as EXR image using Tiny OpenEXR image library, taken from https://github.com/syoyo/tinyexr
 * @param rgb The RGB data to be saved
 * @param width Image width
 * @param height Image height
 * @param outfilename File name for image to be saved as
 */
bool SaveEXR(const float* rgb, int width, int height, const char* outfilename) {

    float* channels[3];

    EXRImage image;
    InitEXRImage(&image);

    image.num_channels = 3;

    // Must be BGR(A) order, since most of EXR viewers expect this channel order.
    const char* channel_names[] = {"B", "G", "R"}; // "B", "G", "R", "A" for RGBA image

    std::vector<float> images[3];
    images[0].resize(width * height);
    images[1].resize(width * height);
    images[2].resize(width * height);

    for (int i = 0; i < width * height; i++) {
      images[0][i] = rgb[3*i+0];
      images[1][i] = rgb[3*i+1];
      images[2][i] = rgb[3*i+2];
    }

    float* image_ptr[3];
    image_ptr[0] = &(images[2].at(0)); // B
    image_ptr[1] = &(images[1].at(0)); // G
    image_ptr[2] = &(images[0].at(0)); // R

    image.channel_names = channel_names;
    image.images = (unsigned char**)image_ptr;
    image.width = width;
    image.height = height;
    image.compression = TINYEXR_COMPRESSIONTYPE_ZIP;

    image.pixel_types = (int *)malloc(sizeof(int) * image.num_channels);
    image.requested_pixel_types = (int *)malloc(sizeof(int) * image.num_channels);
    for (int i = 0; i < image.num_channels; i++) {
      image.pixel_types[i] = TINYEXR_PIXELTYPE_FLOAT; // pixel type of input image
      image.requested_pixel_types[i] = TINYEXR_PIXELTYPE_HALF; // pixel type of output image to be stored in .EXR
    }

    const char* err;
    int ret = SaveMultiChannelEXRToFile(&image, outfilename, &err);
    if (ret != 0) {
      fprintf(stderr, "Save EXR err: %s\n", err);
      return ret;
    }
    //printf("Saved exr file. [ %s ] \n", outfilename);

    free(image.pixel_types);
    free(image.requested_pixel_types);

    return ret;
  }

/**
 * Constructor.
 * @param width Image width.
 * @param height Image height.
 * */
ImageFP::ImageFP(int width, int height)
{
	this->width = width;
	this->height = height;
	this->image = new float [this->width * this->height];
}

/**
 * Destructor.
 * */
ImageFP::~ImageFP()
{
	delete [] this->image;
}

/**
 * Perform two 1D convolutions (horizontal and vertical).
 * @param kernel Array of kernel values of \c kernelSize length.
 * @param kernelSize Size of kernel. Has to be odd.
 * @return 0 on success, 1 on error.
 * */
int ImageFP::convolve1D (float * kernel, int kernelSize)
{
	if(kernelSize % 2 == 0)
	{
		std::cout << "Error in convolve1D: Kernel size even.\n";
		return 1;
	}
	//horizontal
	
	float * im2 = new float [this->height * this->width];	
	#pragma omp parallel for
	for (int i = 0; i < this->height * this->width; ++i)
	{
		float acc = 0;
		for (int j = 0; j < kernelSize; ++j)
		{
			if (i % this->width - floor(kernelSize/2.0) < 0)
			{
				acc += kernel[j] * this->image[i - i % this->width];
			}
			else if (i % this->width + floor(kernelSize/2.0) >= this->width)
			{
				acc += kernel[j] * this->image[i + (this->width - i % this->width)];
			}
			else
			{
				acc += kernel[j] * this->image[i + j - int(floor(kernelSize/2.0))];
			}
		}
		im2[i] = acc;
	}
	delete[] this->image;
	this->image = im2;
	
	//vertical
	
	im2 = new float [this->height * this->width];
	
	for (int i = 0; i < this->height * this->width; ++i)
	{
		float acc = 0;
		for (int j = 0; j < kernelSize; ++j)
		{
			if ((i / this->width) - floor(kernelSize/2.0) < 0)
			{
				acc += kernel[j] * this->image[i - this->width * (i / this->width)];
			}
			else if ((i / this->width) + floor(kernelSize/2.0) >= this->height)
			{
				acc += kernel[j] * this->image[i + this->width * (this->height - i / this->width - 1)];
			}
			else
			{
				acc += kernel[j] * this->image[i + this->width * (j - int(floor(kernelSize/2.0)))];
			}
		}
		im2[i] = acc;
	}
	delete[] this->image;
	this->image = im2;
	
	return 0;
}

/**
 * Add glow to image using kernel, kernel may be applied more times with a custom opacity.
 * @param kernel Array of kernel values of \c kernelSize length.
 * @param kernelSize Size of kernel. Has to be odd.
 * @param howManyTimes How many times the convolution takes place.
 * @param opacity Opacity of the convolution result to be composited into the original.
 * */
void ImageFP::addGlow(float * kernel, int kernelSize, int howManyTimes, float opacity)
{
	float * imageTmp = new float[this->height * this->width];
	for (int i = 0; i<this->height * this->width; ++i)
	{
		imageTmp[i] = this->image[i];
	}
	for (int i = 0; i < howManyTimes; ++i)
	{
		this->convolve1D(kernel, kernelSize);
	}
	for (int i = 0; i<this->height * this->width; ++i)
	{
		if(imageTmp[i] > opacity * this->image[i])
		{
			this->image[i] = imageTmp[i];
		}
		else
		{
			this->image[i] *= opacity;
		}
	}
}