/*
 * P E R U N -- a physically based lightning generator
 * Master's thesis
 * Department of Computer Graphics and Interaction
 * Faculty of Electrical Engineering
 * Czech Technical University in Prague.
 *
 * Supervisor: Ing. Jaroslav Sloup
 *
 * Distributed under the GNU General Public License v3.0
 * */

/**
 * @file 
 * @author Richard Dobrichovsky <dobrichovsky@gmail.com>
 * @date May 2016
 * @brief Definitions for LaplacianGrid class.
 *
 * Definitions for LaplacianGrid class -- the main class which holds all the neccessary methods and variables to simulate the Dielectric Breakdown Model / Laplacian Growth.
 * */

#include "laplacianGrid.h"

/**
 * Constructor
 * @param w Simulation grid width.
 * @param h Simulation grid height.
 * @param generate_new Whether to generate new lightning shape.
 * @param solverType Solver type: SOLVER_EIGEN_REGULAR, SOLVER_EIGEN, SOLVER_PROG2 or SOLVER_CUDA -- 0, 1, 2 or 3 respectively.
 * @param resolutionMultiple Resolution multiplicator.
 * @param frames Number of frames to generate.
 * @param luminance_1 Luminance of main channel.
 * @param luminance_2 Luminance of type 2 branch
 * @param luminance_3 Luminance of other branches.
 * @param type_2_cutoff Minimum length of type 2 branch.
 * @param eta Branching factor: 1 for more branching, 2 for less.
 */
LaplacianGrid::LaplacianGrid(int w, int h,  bool generate_new, int solverType, int resolutionMultiple, int frames,
			double luminance_1,	double luminance_2,	double luminance_3,	int type_2_cutoff, int eta)
{
	this->width = w;
	this->height = h;
	this->total = w * h;
	this->generate_new = generate_new;
	this->solverType = solverType;
	this->resolutionMultiple = resolutionMultiple;
	this->frames = frames;

	this->luminance_1 = luminance_1;
	this->luminance_2 = luminance_2;
	this->luminance_3 = luminance_3;
	this->type_2_cutoff = type_2_cutoff;
	this->eta = eta;

	this->defineGrid();
	for (int i = 0; i < this->width * this->height; ++i)
	{
		this->setImage (i, 0);
		this->setPot (i, 0);
		this->dag[i] = NULL;
		this->probability[i] = 0;
		for (int j = 0; j< this->resolutionMultiple * this->resolutionMultiple; ++j)
		{
			this->imageRender2[i+j] = 0;
		}
	}
	this->printQT = false;
}

/**
 * Destructor
 * */
LaplacianGrid::~LaplacianGrid()
{
	delete[] this->image;
	delete[] this->potential;
	delete[] this->probability;
	delete[] this->dag;
	delete[] this->imageRender2;
}

/**
 * Allocate the working grid
 * @return 0 on success, -1 on error.
 * @exception bad_alloc Bad alloc when unsuccessful in allocating memory for grids
 * */
int LaplacianGrid::defineGrid()
{
	try
	{
		this->image = new unsigned char [this->width * this->height];
	}
	catch (std::bad_alloc & ba)
	{
		std::cerr << "bad_alloc caught in image: " << ba.what() << "\n";
		return -1;
	}
	try
	{
		this->potential = new double [this->width * this->height];
	}
	catch (std::bad_alloc & ba)
	{
		std::cerr << "bad_alloc caught in potential: " << ba.what() << "\n";
		return -1;
	}
	try
	{
		this->probability = new double [this->width * this->height];
	}
	catch (std::bad_alloc & ba)
	{
		std::cerr << "bad_alloc caught in probability: " << ba.what() << "\n";
		return -1;
	}
	try
	{
		this->dag = new DAG*[this->width * this->height];
		for (int i = 0; i < this->width * this->height; ++i)
		{
			this->dag[i] = NULL;
		}
	}
	catch (std::bad_alloc & ba)
	{
		std::cerr << "bad_alloc caught in DAG: " << ba.what() << "\n";
		return -1;
	}
	try
	{
	this->imageRender2 = new unsigned char [this->width * this->resolutionMultiple * this->height *	this->resolutionMultiple];
	}
	catch (std::bad_alloc & ba)
	{
		std::cerr << "bad_alloc caught in imageRender2: " << ba.what() << "\n";
		return -1;
	}
	return 0;
}

/**
 * Origin from where the jolt will spread, with ZERO potential
 * @return 0 on success, -1 on error.
 * */
int LaplacianGrid::setOrigin ()
{
	this->origin.clear();
	for (int i = 0; i < 5; ++i)
	{	
		this->origin.push_back(this->width*i + this->width/2);
		this->setImage(this->width*i + this->width/2, 255);
		if(this->solverType > 0)
		{
			this->quadTree->insert(this->width*i + this->width/2);
		}
		if(i == 0)
		{
			this->dag[this->width*i + this->width/2] = new DAG(this->width*i + this->width/2, NULL,i,this->width/2, this->width, this->height, this->resolutionMultiple);
		}
		else
		{
			this->dag[this->width*i + this->width/2] = new DAG(this->width*i + this->width/2, this->dag[this->width*i + this->width/2 - this->width],i,this->width/2, this->width, this->height, this->resolutionMultiple);
		}
	}
	return 0;
}

/**
 * Set destination, with positive potential (=1)
 * @return 0 on success, -1 on error.
 * */
int LaplacianGrid::setDest ()
{
	this->dest_set.clear();
	int ii = 0;
	//for (int i = this->width * 2.0f / 5.0f; i < this->width * 3.0f / 5.0f; ++i)
	for (int i = this->width * 0.48f; i < this->width * 0.52f; ++i)
	{
		this->dest_set.emplace(this->width * (this->height - 1) + i);
		if(this->solverType > 0)
		{
			this->quadTree->insert(this->width*(this->height-1) + i);
		}
	}
	return 0;
}

/**
 * Set initial charge and find first candidates
 * @return 0 on success, -1 on error.
 * */
int LaplacianGrid::initialiseCandidates()
{
	for (std::vector<int>::iterator it = this->origin.begin(); it != origin.end(); ++it)
	{
		this->grown.emplace (*it);
		this->findCandidates(*it);
	}
	//so that origin sites & borders (only top so far) are not taken in as candidates
	for(std::vector<int>::iterator it = this->origin.begin(); it != origin.end(); ++it)
	{
		if(this->candidates.find(*it) != this->candidates.end())
		{
			this->candidates.erase(this->candidates.find(*it));
		}
		if((*it) < this->width && this->candidates.find(*it + 1) != this->candidates.end())
		{
			this->candidates.erase(this->candidates.find(*it + 1));
		}
		if((*it) < this->width && this->candidates.find(*it - 1) != this->candidates.end())
		{
			this->candidates.erase(this->candidates.find(*it - 1));
		}
	}
	return 0;
}

/**
 * Iteration cycle
 * @return The number of iteration cycles performed.
 * */
int LaplacianGrid::iterate()
{
	int lastbest = -1;
	int farthest = 0;
	int best = this->origin.back();
	this->lastGrown = best;
	int i = 0;
	do
	{
		if(best > farthest)
		{
			farthest = best;
			std::cout << "\r" << 100 * farthest / this->width / this->height << "%";
		}
		findCandidates(best);
		if(i == 0)
		{
			if (this->solverType > 0)
			{
				this->cgSolver->makeAMatrix();
				this->quadTree->subdivided->clear();
			}
			if(this->cgSolver->solvePotential() == -1)
			{
				break;
			}			
		}
		else
		{
			if(this->cgSolver->solvePotential(best) == -1)
			{
				break;
			}			
		}
		this->calculateGrowthProbability();
		lastbest = best;
		best = this->chooseBestCandidate(lastbest);
		this->grow(best);
		this->setImage(best,255);

		//find parent and add to DAG
		//currently ignores jumping accross sides of image but it is improbable that it would be a real problem
		//west
		if(best-1 >=0 && this->image[best - 1] == 255)
		{
			this->dag[best] = new DAG(best, this->dag[best - 1],best/this->width,best%this->width, this->width, this->height, this->resolutionMultiple);			
		}		
		//east
		else if(best+1 < this->width*this->height && this->image[best + 1] == 255)
		{
			this->dag[best] = new DAG(best, this->dag[best + 1],best/this->width,best%this->width, this->width, this->height, this->resolutionMultiple);
		}
		//north
		else if(best-this->width >= 0 && this->image[best - this->width] == 255)
		{
			this->dag[best] = new DAG(best, this->dag[best - this->width],best/this->width,best%this->width, this->width, this->height, this->resolutionMultiple);
		}
		//south
		else if(best+this->width < this->width*this->height && this->image[best + this->width] == 255)
		{
			this->dag[best] = new DAG(best, this->dag[best + this->width],best/this->width,best%this->width, this->width, this->height, this->resolutionMultiple);
		}
		//northwest
		else if(best-this->width-1 > 0 && this->image[best - this->width - 1] == 255)
		{
			this->dag[best] = new DAG(best, this->dag[best - this->width - 1],best/this->width,best%this->width, this->width, this->height, this->resolutionMultiple);
		}
		//northeast
		else if(best-this->width+1 > 0 && this->image[best - this->width + 1] == 255)
		{
			this->dag[best] = new DAG(best, this->dag[best - this->width + 1],best/this->width,best%this->width, this->width, this->height, this->resolutionMultiple);
		}
		//southwest
		else if(best+this->width-1 < this->width*this->height && this->image[best + this->width - 1] == 255)
		{
			this->dag[best] = new DAG(best, this->dag[best + this->width - 1],best/this->width,best%this->width, this->width, this->height, this->resolutionMultiple);
		}
		//southeast
		else if(best+this->width+1 < this->width * this->height && this->image[best + this->width + 1] == 255)
		{
			this->dag[best] = new DAG(best, this->dag[best + this->width + 1],best/this->width,best%this->width, this->width, this->height, this->resolutionMultiple);
		}
		
		++i;
		if(i > ITERATIONS)
		{
			break;
		}
		if(this->grown.size() > GROWLIMIT * this->height)
		{
			break;
		}
	}
	while (this->dest_set.find(best) == this->dest_set.end());
	this->setLuminanace(this->dag[best]);
	//std::cout << "\rDone in " << i << " iterations\n";
	return i;
}

/**
 * Save grid to PPM
 * @param filename Filename
 * @return 0 on success, -1 on error.
 * @todo We don't need this any more.... and this is totally redundant
 * */
int LaplacianGrid::saveImageRender2(const char * filename)
{
	std::ofstream output(filename, std::ofstream::binary|std::ofstream::out);
	if(!output)
	{
		std::cout << " unable to open the output file "<< filename << std::endl;
		return -1;
	}
	else
	{
		output << "P5" << std::endl << "# foreground " << std::endl;
		output << convertInt(width*this->resolutionMultiple) << " ";
		output << convertInt(height*this->resolutionMultiple);
		output << std::endl;
		output << "255" << std::endl;
		output.write( (char *)this->imageRender2, width*height*this->resolutionMultiple*this->resolutionMultiple);
		output.close();
		return 0;
	}
}

/**
 * Method to write integers to string stream for PPM image output
 * @param number Integer to be converted
 * @return String with number
 * */
std::string LaplacianGrid::convertInt(int number)
{
	std::stringstream ss; //create a stringstream
	ss << number; //add number to the stream
	return ss.str(); //return a string with the contents of the stream
}

/**
 * Set potential in grid to value
 * @param index Grid index
 * @param value Grid index potential value
 * @return 0 on success, -1 on error.
 * */
int LaplacianGrid::setPot(int index, double value)
{
	this->potential[index] = value;
	return 0;
}

/**
 * Set pixel in grid to value
 * @param index Image pixel index
 * @param value Pixel value
 * @return 0 on success, -1 on error.
 * */
int LaplacianGrid::setImage(int index, char value)
{
	if(index >= 0)
	{
		this->image[index] = value;
		return 0;
	}
	else
	{
		std::cout << "dafuq? value=" << value << ", index=" << index << "\n";
		return -1;
	}
}

/**
 * Check grid around cand (8 possibilities) and add valid candidates
 * @param cand Chosen candidate
 * @return 0 on success, -1 on error.
 * */
int LaplacianGrid::findCandidates(int cand)
{
	//to prevent jumps from right to left
	if ((cand + 1) % width != 0)
	{
		//right
		if (cand + 1 < total && grown.find(cand + 1) == grown.end())
		{
			this->candidates.emplace (cand + 1);
			if (this->solverType > SOLVER_EIGEN_REGULAR)
			{
				this->quadTree->insert(cand+1);
			}
		}
		//down right
		if (cand + width + 1 < total && grown.find(cand + width + 1) == grown.end())
		{
			this->candidates.emplace (cand + width + 1);
			if (this->solverType > SOLVER_EIGEN_REGULAR)
			{
				this->quadTree->insert(cand + width + 1);
			}
		}
		//up right
		if (cand - width + 1 >= 0 && grown.find(cand - width + 1) == grown.end())
		{
			this->candidates.emplace (cand - width + 1);
			if (this->solverType > SOLVER_EIGEN_REGULAR)
			{				
				this->quadTree->insert(cand - width + 1);
			}			
		}
	}
	//to prevent jumps from left to right
	if ((cand) % width != 0)
	{
		//left
		if (cand - 1 >= 0 && grown.find(cand - 1) == grown.end())
		{
			this->candidates.emplace (cand - 1);
			if (this->solverType > SOLVER_EIGEN_REGULAR)
			{
				this->quadTree->insert(cand - 1);
			}
		}
		//down left
		if (cand + width - 1 < total && grown.find(cand + width - 1) == grown.end() && (cand + width) % width != 0)
		{
			this->candidates.emplace (cand + width - 1);
			if (this->solverType > SOLVER_EIGEN_REGULAR)
			{
				this->quadTree->insert(cand + width - 1);
			}			
		}
		//up left
		if (cand - width - 1 >= 0 && grown.find(cand - width - 1) == grown.end())
		{
			this->candidates.emplace (cand - width - 1);
			if (this->solverType > SOLVER_EIGEN_REGULAR)
			{
				this->quadTree->insert(cand - width - 1);
			}
		}
	}
	//up and down are safe in this respect
	//down
	if (cand + width < total && grown.find(cand + width) == grown.end())
	{
		this->candidates.emplace (cand + width);
		if (this->solverType > SOLVER_EIGEN_REGULAR)
		{
			this->quadTree->insert(cand + width);
		}
	}	
	//up
	if (cand - width >= 0 && grown.find(cand - width) == grown.end())
	{
		this->candidates.emplace (cand - width);
		if (this->solverType > SOLVER_EIGEN_REGULAR)
		{
			this->quadTree->insert(cand - width);
		}
	}
	return 0;
}

/**
 * Calculate probabilities for candidates
 * @return 0 on success, -1 on error.
 * */
int LaplacianGrid::calculateGrowthProbability()
{
	float probability_denom = 0.000001f;
	for (std::unordered_set<int>::iterator it = candidates.begin(); it != this->candidates.end(); ++it)
	{
		probability_denom += pow(this->potential[*it],this->eta);		
	}
	for (std::unordered_set<int>::iterator it = candidates.begin(); it != this->candidates.end(); ++it)
	{
		this->calculateGrowthProbabilityInPoint(*it, probability_denom);
	}
	return 0;
}

/**
 * Calculate probability in point
 * @param point Point index
 * @param probability_denom Probability denominator based on paper
 * @return 0 on success, -1 on error.
 * */
int LaplacianGrid::calculateGrowthProbabilityInPoint(int point, double probability_denom)
{		
	this->probability[point] = pow(this->potential[point],this->eta) / probability_denom;
	return 0;
}

/**
 * Randomly choose candidate based on probabilities
 * @param lastbest Previously chosen candidate
 * @return Chosen candidate
 * */
int LaplacianGrid::chooseBestCandidate(int lastbest)
{
	std::random_device rd;
	std::default_random_engine el(rd());
	int size = candidates.size();
	std::uniform_real_distribution<float> uniform_dist(0,1);
	float chosen = uniform_dist(el);

	//float best = this->pot_min;
	int best_pointer = -1;
	float probability_accumulator = 0;
	int prev;
	int preprev;	
	for (std::unordered_set<int>::iterator it = candidates.begin(); it != this->candidates.end(); ++it)
	{
		if (chosen > probability_accumulator)
		{
			probability_accumulator += this->probability[*it];
			prev = *it;
		}
		else
		{
			return prev;
		}
	}
	return prev;
}

/**
 * Add chosen candidate to bolt and erase from candidates
 * @param best Chosen candidate
 * @return 0 on success, -1 on error.
 * */
int LaplacianGrid::grow(int best)
{
	grown.emplace(best);
	candidates.erase(best);	
	this->lastGrown = best;
	return 0;
}

/**
 * Initialise Conjugate Gradient solver
 * */
void LaplacianGrid::initSolver()
{
	//float ttolerance = 0.0005;
	double ttolerance = 0.00001;
	
	double bias = this->height / 128.0f;
	bias *= bias;
	ttolerance /= bias;
	if(this->solverType > SOLVER_EIGEN_REGULAR)
	{
		this->cgSolver = new CGSolver(this->width, this->height, this->quadTree, &this->grown, &this->grown_last, &this->dest_set, this->potential,1000,ttolerance, this->solverType);
	}
	else
	{
		this->cgSolver = new CGSolver(this->width, this->height, &this->grown, &this->dest_set, this->potential,1000,ttolerance);
	}
}

/**
 * Decimal part of number used in Xiaolin Wu's line algorithm
 * @param x Input number
 * @return Decimal part of \a x
 * */
float fpart (float x)
{
	if(x<0)
	{
		return 1 - (x - floor(x));
	}
	return x - floor(x);
}

/**
 * 1-\c fpart(x)
 * @param x Imput number
 * @return 1 - decimal part of \a x
 * */
float rfpart (float x)
{
	return 1 - fpart(x);
}

/**
 * \a x rounded
 * @param x Input number
 * @return Input number rounded
 * */
float round(float x)
{
	return int(x + 0.5);
}

/**
* Xiaolin Wu's line algorithm
* https://en.wikipedia.org/wiki/Xiaolin_Wu's_line_algorithm
* @param a Line starting point
* @param b Line endpoint
*/
void LaplacianGrid::drawLine2(DAG * a, DAG * b)
{
	float luminanceA = a->luminance;
	float luminanceB = b->luminance;
	if(a->type != b->type)
	{
		luminanceA = luminanceB;
	}
	float luminance = 0.0f;
	int rWIDTH = this->width * this->resolutionMultiple;
	int rHEIGHT = this->height * this->resolutionMultiple;
	bool steep = abs(b->y - a->y) > abs(b->x - a->x);
	float x0,x1,y0,y1;
	if (steep)
	{
		x0 = a->y;
		x1 = b->y;
		y0 = a->x;
		y1 = b->x;
	}
	else
	{
		x0 = a->x;
		x1 = b->x;
		y0 = a->y;
		y1 = b->y;
	}
	if (x0 > x1)
	{
		float t = x1;
		x1 = x0;
		x0 = t;
		t = y1;
		y1 = y0;
		y0 = t;
		t = luminanceB;
		luminanceB = luminanceA;
		luminanceA = t;
	}


	float dx = x1 - x0;
	float dy = y1 - y0;
	float gradient  = dy / dx;

	//first endpoint
	float xend = round(x0);
	float yend = y0 + gradient * (xend - x0);
	float xgap = rfpart(x0 + 0.5);
	int xpxl1 = (int)xend;
	int ypxl1 = (int)yend;
	if(steep)
	{
		//this->imageFP->image[(X) + rWIDTH * (Y)];
		if(this->imageFP->image[(int)((ypxl1) + rWIDTH * (xpxl1))] < rfpart(yend) * xgap * luminanceA)
		{
			this->imageFP->image[(int)((ypxl1) + rWIDTH * (xpxl1))] = rfpart(yend) * xgap * luminanceA;
		}
		if(this->imageFP->image[(int)((ypxl1 + 1) + rWIDTH * (xpxl1))] < fpart(yend) * xgap * luminanceA)
		{
			this->imageFP->image[(int)((ypxl1 + 1) + rWIDTH * (xpxl1))] = fpart(yend) * xgap * luminanceA;
		}
	}
	else
	{
		if(this->imageFP->image[(int)((xpxl1) + rWIDTH * (ypxl1))] < rfpart(yend) * xgap * luminanceA)
		{
			this->imageFP->image[(int)((xpxl1) + rWIDTH * (ypxl1))] = rfpart(yend) * xgap * luminanceA;
		}
		if(this->imageFP->image[(int)((xpxl1) + rWIDTH * (ypxl1 + 1))] < fpart(yend) * xgap * luminanceA)
		{
			this->imageFP->image[(int)((xpxl1) + rWIDTH * (ypxl1 + 1))] = fpart(yend) * xgap * luminanceA;
		}
	}
	float intery = yend + gradient;

	// second endpoint
	xend = round(x1);
	yend = y1 + gradient * (xend - x1);
	xgap = fpart(x1 + 0.5);
	int xpxl2 = xend;
	int ypxl2 = (int)yend;
	if(steep)
	{
		//this->imageFP->image[(X) + rWIDTH * (Y)];
		if(this->imageFP->image[(int)((ypxl2) + rWIDTH * (xpxl2))] < rfpart(yend) * xgap * luminanceB)
		{
			this->imageFP->image[(int)((ypxl2) + rWIDTH * (xpxl2))] = rfpart(yend) * xgap * luminanceB;
		}
		if(this->imageFP->image[(int)((ypxl2 + 1) + rWIDTH * (xpxl2))] < fpart(yend) * xgap * luminanceB)
		{
			this->imageFP->image[(int)((ypxl2 + 1) + rWIDTH * (xpxl2))] = fpart(yend) * xgap * luminanceB;
		}
	}
	else
	{
		if(this->imageFP->image[(int)((xpxl2) + rWIDTH * (ypxl2))] < rfpart(yend) * xgap * luminanceB)
		{
			this->imageFP->image[(int)((xpxl2) + rWIDTH * (ypxl2))] = rfpart(yend) * xgap * luminanceB;

		}
		if(this->imageFP->image[(int)((xpxl2) + rWIDTH * (ypxl2 + 1))] < fpart(yend) * xgap * luminanceB)
		{
			this->imageFP->image[(int)((xpxl2) + rWIDTH * (ypxl2 + 1))] = fpart(yend) * xgap * luminanceB;
		}
	}
	float luminanceChange = (luminanceA - luminanceB) / (xpxl2-(xpxl1+1));
	luminance = luminanceA;
	for (int x = xpxl1+1; x<xpxl2; ++x)
	{
		luminance -= luminanceChange;
		if(steep)
		{
			//this->imageFP->image[(X) + rWIDTH * (Y)];
			if(this->imageFP->image[(int)(((int) intery) + rWIDTH * x)] < rfpart(intery) * luminance)
			{
				this->imageFP->image[(int)(((int) intery) + rWIDTH * x)] = rfpart(intery) * luminance;
			}
			if(this->imageFP->image[(int)(((int) intery+1) + rWIDTH * x)] < fpart(intery) * luminance)
			{
				this->imageFP->image[(int)(((int) intery+1) + rWIDTH * x)] = fpart(intery) * luminance;
			}
		}
		else
		{
			if(this->imageFP->image[(int)(((int) intery)*rWIDTH + x)] < rfpart(intery) * luminance)
			{
				this->imageFP->image[(int)(((int) intery)*rWIDTH + x)] = rfpart(intery) * luminance;
			}
			if(this->imageFP->image[(int)(((int) intery+1)*rWIDTH + x)] < fpart(intery) * luminance)
			{
				this->imageFP->image[(int)(((int) intery+1)*rWIDTH + x)] = fpart(intery) * luminance;
			}
		}
		intery += gradient;
	}
}

/**
 * Recursively traverse through DAG and draw lines between nodes
 * @param dag Parent node
 * */
void LaplacianGrid::DAG_traverse(DAG * dag)
{
	for(std::vector<DAG*>::iterator it = dag->children.begin(); it != dag->children.end(); ++it)
	{
		this->drawLine2(dag,*it);
		DAG_traverse(*it);		
	}
}

/**
 * Traverse through DAG from terminal node (where the lightning hit the ground) to parent setting the main channel to a higher luminance value
 * @param dag Terminal node
 * */
void LaplacianGrid::setLuminanace(DAG * dag)
{
	dag->luminance = this->luminance_1;
	dag->type = 1;
	if(dag->parent != NULL)
	{
		this->setLuminanace(dag->parent);
	}
}

/**
 * Traverse along main channel and run \c DAG_findBranchesSide() to find secondary and ternary branches
 * @param dag Start node
 * */
void LaplacianGrid::DAG_findBranchesMain(DAG * dag)
{
	DAG * next = NULL;
	std::list<DAG*> * bestList = NULL;
	std::list<DAG*> * tempList = NULL;
	for(std::vector<DAG*>::iterator it = dag->children.begin(); it != dag->children.end(); ++it)
	{
		if((*it)->type == 1)
		{
			(*it)->luminance = this->luminance_1;
			next = *it;			
		}
		else
		{
			if(bestList == NULL)
			{
				bestList = DAG_findBranchesSide(*it);
			}
			else
			{
				tempList = DAG_findBranchesSide(*it);
				if(tempList->size() > bestList->size())
				{
					//we can't do this now because we don't know the intensity at the root
					//this->DAG_setBranchMinor(bestList);
					//delete bestList;
					this->listAccum.push_back(bestList);
					bestList = tempList;
				}
				else
				{
					this->listAccum.push_back(tempList);
					//we can't do this now because we don't know the intensity at the root
					//this->DAG_setBranchMinor(tempList);
				}
			}
		}
	}
	if(bestList != NULL)
	{
		//We have found the longest branch; if it qualifies to be a type 2 branch, 
		if(bestList->size() > this->type_2_cutoff)
		{
			this->DAG_setType2Branch(bestList);
		}
		else
		{
			this->DAG_setType3Branch(bestList);
		}
		delete bestList;
	}
	if (next != NULL)
	{
		DAG_findBranchesMain(next);
	}
	else
	{
		return;
	}
}

/**
 * Traverse through side branches and find the longest one
 * @param dag Start node
 * @return The longest branch
 * */
std::list<DAG*> * LaplacianGrid::DAG_findBranchesSide(DAG * dag)
{
	std::list<DAG*> * theList = new std::list<DAG*>();
	theList->push_back(dag);
	while (dag->children.size() == 1)
	{
		dag = *dag->children.begin();
		theList->push_back(dag);
	}
	if (dag->children.size() == 0)
	{
		return theList;
	}
	std::list<DAG*> * bestList = NULL;
	std::list<DAG*> * tempList = NULL;
	for (std::vector<DAG*>::iterator it = dag->children.begin(); it != dag->children.end(); ++it)
	{
		if(bestList == NULL)
		{
			bestList = DAG_findBranchesSide(*it);
		}
		else
		{
			tempList = DAG_findBranchesSide(*it);
			if(tempList->size() > bestList->size())
			{
				//delete bestList;
				this->listAccum.push_back(bestList);
				bestList = tempList;
			}
			else
			{
				this->listAccum.push_back(tempList);
			}
		}
	}
	theList->splice(theList->end(),*bestList);
	delete bestList;
	return theList;
}

/**
 * Sets nodes to type 2 and appropriate intensity
 * @todo custom fallof, custom fallof scheme (now only linear -- log? something else "smooth"?)
 * @param theList List of graph nodes to set type and intensity
 * */
void LaplacianGrid::DAG_setType2Branch(std::list<DAG*> * theList)
{
	float coeffiecient = 1.0f * this->luminance_2 /theList->size();
	int i = 1;
	for(std::list<DAG*>::iterator it = theList->begin(); it != theList->end(); ++it)
	{
		if((*it)->type != 1)
		{
			(*it)->type = 2;
			(*it)->luminance = this->luminance_2 - i * coeffiecient;
		}
		++i;
	}
}

/**
 * Sets nodes to type 3 and intensity based on the intensity of the root's parent
 * @todo custom fallof, custom fallof scheme (now only linear -- log? something else "smooth"?)
 * @param theList List of graph nodes to set type and intensity
 * */
void LaplacianGrid::DAG_setType3Branch(std::list<DAG*> * theList)
{
	float luminanceMax = 0.0f;
	std::list<DAG*>::iterator it = theList->begin();
	if (it != theList->end())
	{
		if((*it)->parent->type != 1)
		{
			luminanceMax = std::min((*it)->parent->luminance,this->luminance_3);
		}
		else
		{
			luminanceMax = this->luminance_3;
		}
	}
	else
	{
		return;
	}
	float coefficient = luminanceMax/theList->size();
	int i = 1;
	for(it; it != theList->end(); ++it)
	{
		if((*it)->type != 1 && (*it)->type != 2)
		{
			(*it)->luminance = luminanceMax - i * coefficient;
			(*it)->type = 3;
			//(*it)->luminance = 0;
			++i;
		}
	}
}

/**
 * Starting from longest list in \c listAccum runs \c DAG_setType3Branch() on each
 * @todo Proper sorting (but it works like this)
 * */
void LaplacianGrid::DAG_setType3Lists()
{
	//How the fuck does this know how to sort? But it does the job.
	std::sort(this->listAccum.begin(),this->listAccum.end());
	
	for (std::vector<std::list<DAG*>*>::iterator it = this->listAccum.begin(); it != this->listAccum.end(); ++it)
	{
		if(*it != NULL)
		{
			this->DAG_setType3Branch(*it);
			delete *it;
			*it = NULL;
		}
	}
	this->listAccum.clear();
}

/**
 * Load node graph from file.
 * @param frame Frame number.
 * */
void LaplacianGrid::loadGraph(int frame)
{
	if(this->dag[this->width / 2] != NULL)
	{
		delete this->dag[this->width / 2];
	}
	
	std::string fname = "lightning_graph_";
	if (frame < 10)
	{
		fname += "000";		
	}
	else if (frame < 100)
	{
		fname += "00";
	}
	else if (frame < 1000)
	{
		fname += "0";
	}
	std::string fnumber = std::to_string(static_cast<long long>(frame));
	fname += fnumber;	

	std::ifstream input (fname);
	std::string tmp;
	input >> tmp;
	if(tmp.compare("vertices") != 0)
	{
		std::cout << "Syntax error in " << fname << ": vertices\n";
		return;
	}
	int id;
	float x;
	float y;
	int type;
	float luminance;
	
	while(true)
	{
		input >> id;
		if(id != -1)
		{
			input >> x;
			input >> y;
			input >> type;
			input >> luminance;
			this->dag[id] = new DAG(id, x, y, type, luminance);
		}
		else
		{
			break;
		}
	}
	input >> tmp;
	if(tmp.compare("edges") != 0)
	{
		std::cout << "Syntax error in " << fname << ": vertices\n";
		return;
	}
	int parent;
	int child;
	while (true)
	{
		input >> parent;
		if(parent != -1)
		{
			input >> child;
			this->dag[parent]->children.push_back(this->dag[child]);
			this->dag[child]->parent = this->dag[parent];
		}
		else
		{
			break;
		}
	}
}

/**
 * Reset internal variables as preparation for new frame.
 * */
void LaplacianGrid::newFrame()
{
	delete this->imageFP;
	delete this->cgSolver;
	this->grown_last.clear();
	for(std::unordered_set<int>::iterator it = this->grown.begin(); it != this->grown.end(); ++it)
	{
		if(this->dag[*it] != NULL && this->dag[*it]->type == 1)
		{
			this->grown_last.emplace(*it);
		}
	}
	this->grown.clear();
	this->dest_set.clear();
	this->candidates.clear();
	delete this->dag[this->width/2];
	for (int i = 0; i < this->width * this->height; ++i)
	{
		this->dag[i] = NULL;
		this->image[i] = 0;
	}
}

/**
 * Save EXR image of rendered frame.
 * @param frame Frame number.
 * */
void LaplacianGrid::saveEXR(int frame)
{
	float * exrdata = new float[this->imageFP->width * this->imageFP->height * 3];
	for(int i = 0; i<this->imageFP->width * this->imageFP->height; ++i)
	{
		exrdata[3*i] = this->imageFP->image[i];
		exrdata[3*i+1] = this->imageFP->image[i];
		exrdata[3*i+2] = this->imageFP->image[i];
	}
	std::string fname = "bolt_render_conv_";
	if (frame < 10)
	{
		fname += "000";		
	}
	else if (frame < 100)
	{
		fname += "00";
	}
	else if (frame < 1000)
	{
		fname += "0";
	}
	std::string fnumber = std::to_string(static_cast<long long>(frame));
	fname += fnumber;
	fname += ".exr";	
	SaveEXR(exrdata,this->imageFP->width, this->imageFP->height,fname.c_str());
	delete[] exrdata;
}

/**
 * Perform two 1D convolutions over a Gaussian kernel of size 7.
 * */
void LaplacianGrid::addGlow()
{
	float * u2 = new float [7];
	u2[0] = 2.107793930598946 * 0.00038771;
	u2[1] = 2.107793930598946 * 0.01330373;
	u2[2] = 2.107793930598946 * 0.11098164;
	u2[3] = 2.107793930598946 * 0.22508352;
	u2[4] = 2.107793930598946 * 0.11098164;
	u2[5] = 2.107793930598946 * 0.01330373;
	u2[6] = 2.107793930598946 * 0.00038771;
	this->imageFP->convolve1D(u2,7);
	delete[] u2;
}

/**
 * Run the simulation.
 * */
void LaplacianGrid::run()
{
	std::cout << "P E R U N -- a physically based lightning generator\n";
	std::cout << this->width << "x" << this->height << ", eta=" << this->eta;
	if(this->solverType == SOLVER_EIGEN_REGULAR)
	{
		std::cout << ", regular";
	}
	else if(this->solverType == SOLVER_EIGEN)
	{
		std::cout << ", eigen";
	}
	else if(this->solverType == SOLVER_PROG2)
	{
		std::cout << ", prog2";
	}
	else
	{
		std::cout << ", cuda";
	}
	std::cout << "\n";
	clock_t timer3, timer4;
	timer3 = clock();
	int totaltotaliterations = 0;
	
	if(this->solverType > SOLVER_EIGEN_REGULAR)
	{
		this->quadTree = new QuadTree(this->width, this->height, LEVEL, this->resolutionMultiple);
	}
	for(int frame = 1; frame <= this->frames; ++frame)
	{
		clock_t timer1, timer2;
		timer1 = clock();

		this->imageFP = new ImageFP(this->width * this->resolutionMultiple, this->height * this->resolutionMultiple);
		

		//this->dag[this->width/2] = new DAG(this->width/2,NULL,0,this->width/2);
	
		
		int totaliterations = 0;
		if(this->generate_new)
		{
			this->setDest();
			this->setOrigin();
	
			this->initialiseCandidates();
			this->initSolver();
		
			totaliterations = this->iterate();
		}

		if(this->printQT && frame == this->frames)
		{
			this->quadTree->printQuadTree(this->imageRender2);
			if(this->saveImageRender2("quadtree.ppm") == 0)
			{
				std::cout << "Saved [quadtree.ppm]\n";
			}
			else
			{
				std::cout << "Error in saving [quadtree.ppm]\n";
			}
			this->quadTree->memory();
		}
				
		if(!this->generate_new)
		{
			this->loadGraph(frame);
		}

		this->DAG_findBranchesMain(this->dag[this->width/2]);
		this->DAG_setType3Lists();
		this->DAG_traverse(this->dag[this->width/2]);
		timer2 = clock();


		if(this->generate_new)
		{
			this->dag[this->width/2]->saveGraph(frame);
		

			

		
	#ifdef WIN32
			std::cout << "\rTotal iterations: " << totaliterations << "\n";
			std::cout << "Time taken: " << (timer2 - timer1) << "ms\n";		
			std::cout << "Time per iteration: " << 1.0f * ((timer2 - timer1)) / totaliterations << "ms\n";
	#else
			std::cout << "\rTotal iterations: " << totaliterations << "\n";
			std::cout << "Time taken: " << (timer2 - timer1) / 1000.0f << "ms\n";	
			std::cout << "Time per iteration: " << 1.0f * ((timer2 - timer1)) / 1000.0f / totaliterations << "ms\n";
	#endif
			std::cout << "-------------------\n";
			
		}

		this->addGlow();
		this->saveEXR(frame);

		this->newFrame();
		
		totaltotaliterations += totaliterations;
		++frameno;
		
	}

	if(this->solverType > SOLVER_EIGEN_REGULAR)
	{
		delete this->quadTree;
	}

	
	if(this->generate_new)
	{
		timer4 = clock();
	#ifdef WIN32
		std::cout << "\rTotal time taken: " << (timer4 - timer3) << "ms\n";		
		std::cout << "Total iterations: " << totaltotaliterations << "\n";		
		std::cout << "Average time per frame: " << 1.0f * ((timer4 - timer3)) / this->frames << "ms\n";
		std::cout << "Average time per iteration: " << 1.0f * ((timer4 - timer3)) / totaltotaliterations << "ms\n";
	#else
		std::cout << "\rTotal time taken: " << (timer4 - timer3) / 1000.0f << "ms\n";		
		std::cout << "Total iterations: " << totaltotaliterations << "\n";		
		std::cout << "Average time per frame: " << 1.0f * ((timer4 - timer3)) / 1000.0f / this->frames << "ms\n";
		std::cout << "Average time per iteration: " << 1.0f * ((timer4 - timer3)) / 1000.0f / totaltotaliterations << "ms\n";
	#endif
		std::cout << "-------------------\n";
	}
	else
	{
		std::cout << "Image edit finished.\n";
	}
}