/*
 * P E R U N -- a physically based lightning generator
 * Master's thesis
 * Department of Computer Graphics and Interaction
 * Faculty of Electrical Engineering
 * Czech Technical University in Prague.
 *
 * Supervisor: Ing. Jaroslav Sloup
 *
 * Distributed under the GNU General Public License v3.0
 * */

/**
 * @file laplacianGrid.h
 * @author Richard Dobrichovsky <dobrichovsky@gmail.com>
 * @date May 2016
 * @brief Declarations for LaplacianGrid class.
 *
 * Declarations for LaplacianGrid class -- the main class which holds all the neccessary methods and variables to simulate the Dielectric Breakdown Model / Laplacian Growth.
 * */

#include <string>
#include <list>
#include <unordered_set>

#include "graph.h"
#include "imageFP.h"
#include "cg_solver.h"


/**
 * Main class which holds all the neccessary methods and variables to simulate the Dielectric Breakdown Model / Laplacian Growth 
 * @brief Main class which holds all the neccessary methods and variables to simulate the Dielectric Breakdown Model / Laplacian Growth 
 * */
class LaplacianGrid
{
public:
	LaplacianGrid (int width, int height, bool generate_new, int solverType, int resolutionMultiple, int frames,
			double luminance_1,	double luminance_2,	double luminance_3,	int type_2_cutoff, int eta);
	~LaplacianGrid ();
	void run();
private:
	int defineGrid ();
	int setOrigin ();
	int setDest ();
	int initialiseCandidates();
	void initSolver();
	int iterate();
	int saveImageRender2 (const char * filename);
	std::string convertInt(int number);	
	void saveEXR(int frame);
	void drawLine2(DAG * a, DAG * b);
	void DAG_traverse(DAG * dag);
	void setLuminanace(DAG * dag);
	void DAG_findBranchesMain(DAG * dag);
	std::list<DAG*> * DAG_findBranchesSide(DAG * dag);
	void DAG_setType2Branch(std::list<DAG*> * theList);
	void DAG_setType3Branch(std::list<DAG*> * theList);
	void DAG_setType3Lists();
	void loadGraph(int frame);
	void addGlow();
	void newFrame();
	int setImage (int index, char value);
	int setPot (int index, double value);
	int findCandidates(int cand);
	int updatePotential();	
	int calculateGrowthProbability();
	int calculateGrowthProbabilityInPoint(int point, double probability_denom);	
	int chooseBestCandidate(int lastbest);		
	int grow(int best);

	/// Array of pointers to DAG nodes
	DAG ** dag;
	/// Image to render quadtree if desired
	unsigned char * imageRender2;
	/// Floating point image for rendering
	ImageFP * imageFP;
	/// List accumulator used in DAG operations
	std::vector<std::list<DAG*>*> listAccum;
	/// Quadtree for adaptive grid
	QuadTree * quadTree;
	/// Conjugate gradient solver
	CGSolver * cgSolver;
	/// Grid width
	int width;
	/// Grid height
	int height;
	/// Resolution Multiplicator
	int resolutionMultiple;
	/// Total number of frames to render
	int frames;
	/// Whether to generate a new shape or load from file
	bool generate_new;
	/// Whether to render quadtree
	bool printQT;
	/// Solver type
	int solverType;
	/// Luminance of main channel
	double luminance_1;
	/// Luminance of type 2 branch
	double luminance_2;
	/// Luminance of other branches
	double luminance_3;
	/// Minimum length of type 2 branch
	int type_2_cutoff;
	/// Eta -- branching factor
	int eta;		
	/// Index of last grown element
	int lastGrown;
	/// Grid of potential values	
	double * potential;
	/// Grid of probability values
	double * probability;
	/// Resulting image
	unsigned char * image;

	/// Grid width * grid height
	int total;
	/// Vector of origin sites
	std::vector<int> origin;
	/// Set of destination sites for more efficient find operations
	std::unordered_set<int> dest_set;
	/// Set of candidate sites
	std::unordered_set<int> candidates;
	/// Set of growth sites
	std::unordered_set<int> grown;
	/// Set pf growth sites in previous frame
	std::unordered_set<int> grown_last;	
};
