/*
 * P E R U N -- a physically based lightning generator
 * Master's thesis
 * Department of Computer Graphics and Interaction
 * Faculty of Electrical Engineering
 * Czech Technical University in Prague.
 *
 * Supervisor: Ing. Jaroslav Sloup
 *
 * Distributed under the GNU General Public License v3.0
 * */

/**
 * @file
 * @author Richard Dobrichovsky <dobrichovsky@gmail.com>
 * @date May 2016
 * 
 * @brief Main project file.
 *
 * Implements main function that parses command line options and runs the Laplacian grid simulation.
 *
 * */

#include <string>
#include "common.h"
#include "perun.h"
#include "laplacianGrid.h"

/**
 * Main function. Parses command line arguments and starts the simulation.
 * */
int main (int argc, char** argv)
{
	int width = 64;
	int height = 128;
	bool generate_new = true;
	bool solverExplicit = false;
	int solverType = SOLVER_PROG2;
	int resolutionMultiple = 8;
	int frames = 1;
	double l1 = 4.0;
	double l2 = 1.0;
	double l3 = 0.1;
	int t2 = 20;
	int eta = 2;

	for (int i = 1; i < argc; ++i)
	{
		if (std::string(argv[i]) == "--help")
		{
			std::cout << "P E R U N -- a physically based lightning generator\n";
			std::cout << " Command line options:\n";
			std::cout << "  -e ETA            Branching factor. ETA = 1 more branching, ETA = 2 less\n                     branching.\n";
			std::cout << "  -f FRAMES         Number of frames to animate.\n";
			std::cout << "  -g                Do not generate new shape.\n";
			std::cout << "  -h HEIGHT         Simulation grid height.\n";
			std::cout << "  -l1 LUMINANCE     Luminance of main channel. Default = 4.0.\n";
			std::cout << "  -l2 LUMINANCE     Luminance of secondary channels. Secondary channels are\n                    those longer than TYPE_2_CUTTOFF. Default LUMINANCE = 1.0.\n";
			std::cout << "  -l3 LUMINANCE     Luminance of ternary channels. Ternary channels are all\n                    others.\n";
			std::cout << "  -r RESOLUTION     Resolution multiplicator of rendered image. Resulting\n                    image width will be WIDTH * RESOLUTION. Default\n                    RESOLUTION = 8.\n";
			std::cout << "  -s SOLVER         Solver to use. Valid options regular, eigen, prog2 or cuda.\n";
			std::cout << "  -t2 TYPE_2_CUTOFF Secondary channels cuttof. Default TYPE_2_CUTOFF = 20.\n";
			std::cout << "  -w WIDTH          Simulation grid width.\n";
			return 0;
		}
		else if (std::string(argv[i]) == "-w")
		{
			if (i + 1 < argc)
			{
				width = std::stoi(argv[++i]);
			}
			else
			{
				std::cerr << "-w option requires one argument." << std::endl;
				return 1;
			}
		}
		else if (std::string(argv[i]) == "-h")
		{
			if (i + 1 < argc)
			{
				height = std::stoi(argv[++i]);
			}
			else
			{
				std::cerr << "-h option requires one argument." << std::endl;
				return 1;
			}
		}
		else if (std::string(argv[i]) == "-g")
		{
			generate_new = false;
		}
		else if (std::string(argv[i]) == "-s")
		{
			if (i + 1 < argc)
			{
				solverExplicit = true;
				++i;
				if (std::string(argv[i]) == "regular")
				{
					solverType = SOLVER_EIGEN_REGULAR;
				}
				else if (std::string(argv[i]) == "eigen")
				{
					solverType = SOLVER_EIGEN;
				}
				else if (std::string(argv[i]) == "prog2")
				{
					solverType = SOLVER_PROG2;
				}
				else if (std::string(argv[i]) == "cuda")
				{
					solverType = SOLVER_CUDA;
				}
				else
				{
					std::cerr << "-s: invlaid argument." << std::endl;
					return 1;
				}
			}
			else
			{
				std::cerr << "-s option requires one argument." << std::endl;
				return 1;
			}
		}
		else if (std::string(argv[i]) == "-r")
		{
			if (i + 1 < argc)
			{
				resolutionMultiple = std::stoi(argv[++i]);
			}
			else
			{
				std::cerr << "-r option requires one argument." << std::endl;
				return 1;
			}
		}
		else if (std::string(argv[i]) == "-f")
		{
			if (i + 1 < argc)
			{
				frames = std::stoi(argv[++i]);
			}
			else
			{
				std::cerr << "-f option requires one argument." << std::endl;
				return 1;
			}
		}
		else if (std::string(argv[i]) == "-l1")
		{
			if (i + 1 < argc)
			{
				l1 = std::stod(argv[++i]);
			}
			else
			{
				std::cerr << "-l1 option requires one argument." << std::endl;
				return 1;
			}
		}
		else if (std::string(argv[i]) == "-l2")
		{
			if (i + 1 < argc)
			{
				l2 = std::stod(argv[++i]);
			}
			else
			{
				std::cerr << "-l2 option requires one argument." << std::endl;
				return 1;
			}
		}
		else if (std::string(argv[i]) == "-l3")
		{
			if (i + 1 < argc)
			{
				l3 = std::stod(argv[++i]);
			}
			else
			{
				std::cerr << "-l3 option requires one argument." << std::endl;
				return 1;
			}
		}
		else if (std::string(argv[i]) == "-t2")
		{
			if (i + 1 < argc)
			{
				t2 = std::stoi(argv[++i]);
			}
			else
			{
				std::cerr << "-t2 option requires one argument." << std::endl;
				return 1;
			}
		}
		else if (std::string(argv[i]) == "-e")
		{
			if (i + 1 < argc)
			{
				eta = std::stoi(argv[++i]);
			}
			else
			{
				std::cerr << "-e option requires one argument." << std::endl;
				return 1;
			}
		}
	}

	if(width != height && 2*width != height)
	{
		std::cerr << "Invalid resolution." << std::endl;
		return 1;
	}

	if(!solverExplicit)
	{
		if(height >= 512)
		{
			solverType = SOLVER_CUDA;
		}
	}

	if(eta != 1 && eta != 2)
	{
		std::cerr << "Eta invalid." << std::endl;
		return 1;
	}

	LaplacianGrid * laplacianGrid = new LaplacianGrid(width, height, generate_new, solverType, resolutionMultiple, frames,
		l1, l2, l3, t2, eta);

	//LaplacianGrid * laplacianGrid = new LaplacianGrid(64, 128, false, SOLVER_PROG2, 8, 20,
	//	3.0, 1.0, 0.1, 20, 2);
	
	laplacianGrid->run();
	
	delete laplacianGrid;
}