/*
 * P E R U N -- a physically based lightning generator
 * Master's thesis
 * Department of Computer Graphics and Interaction
 * Faculty of Electrical Engineering
 * Czech Technical University in Prague.
 *
 * Supervisor: Ing. Jaroslav Sloup
 *
 * Distributed under the GNU General Public License v3.0
 * */

/**
 * @file
 * @author Richard Dobrichovsky <dobrichovsky@gmail.com>
 * @date May 2016
 * 
 * @brief Definitions for QuadTree and QuadTreeNode classes.
 *
 * Definitions for QuadTree and QuadTreeNode classes implementing a quadtree data structure used in the adaptive grid
 * versions of conjugate gradients solvers.
 *
 * */

#include "common.h"
#include "quadtree.h"

/**
 * Constructor
 * @param width Original grid resolution width in pixels.
 * @param height Original grid resolution height in pixels.
 * @param minLevel Minimum level of quadtree leaf nodes.
 * @param resolutionMultiple Resolution multiplicator.
 * @todo Make sure sizes are compatible
 * */
QuadTree::QuadTree(int width, int height, int minLevel, int resolutionMultiple)
{
	this->width = width;
	this->height = height;
	this->minLevel = minLevel;	
	this->resolutionMultiple = resolutionMultiple;
	this->leaves = new std::set<QuadTreeNode *>();
	this->subdivided = new std::set<QuadTreeNode *>();
	
	if(this->height > this->width)
	{
		this->root = new QuadTreeNode(NULL, 0, _R, 0, this->height, this);
		this->root->NW = new QuadTreeNode(this->root, 1, 3, 0, this->width, this);
		this->leaves->insert(this->root->NW);
		this->subdivided->insert(this->root->NW);
		this->root->SW = new QuadTreeNode(this->root, 1, 1, this->width * this->height/2, this->width, this);
		this->leaves->insert(this->root->SW);
		this->subdivided->insert(this->root->SW);
	}
	else
	{
		this->root = new QuadTreeNode(NULL, 0, _R, 0, this->width, this);
	}
	this->root->subdivide(minLevel);
}

/**
 * Destructor. Destroys leaves and subdivided objects as well as the quadtree node hierarchy from root node.
 * */
QuadTree::~QuadTree()
{
	delete this->leaves;
	delete this->subdivided;
	delete this->root;
}

/**
 * Visualise quadtee.
 * @param image Array of pixels to write to. Has to be WIDTH * HEIGHT * RESOLUTION * RESOLUTION
 * */
void QuadTree::printQuadTree(unsigned char * image)
{
	for (std::set<QuadTreeNode *>::iterator it = this->leaves->begin(); it != this->leaves->end(); ++it)
	{
		(*it)->printQuadTreeNode(image);
	}
}

/**
 * Insert pixel into quadree: the pixel will be at full resolution
 * @param pixel Pixel coordinate
 * @return Pointer to quadtree node corresponding to \a pixel
 * */
QuadTreeNode * QuadTree::insert(int pixel)
{
	QuadTreeNode * currentNode = this->root;
	int x = pixel % this->width;
	int y = pixel / this->width;
	int cutX;
	int cutY;
	while (currentNode->size > 1)
	{
		if(currentNode == this->root && this->height > this->width)
		{
			cutY = this->height / 2;
			cutX = this->width;
		}
		else
		{
			cutX = currentNode->topLeft % this->width + currentNode->size/2;
			cutY = currentNode->topLeft / this->width + currentNode->size/2;
		}
		if (currentNode->isLeaf())
		{
			//this->subdivided->erase(currentNode);
			currentNode->subdivide();
			/*
			std::vector<QuadTreeNode*> neighbours = currentNode->getAllNeighbours();
			for (std::vector<QuadTreeNode*>::iterator it = neighbours.begin(); it != neighbours.end(); ++it)
			{
				if ((*it) != NULL && (*it)->level < currentNode->level)
				{
					(*it)->subdivide();
				}
			}*/
		}
		if (x < cutX)			
		{
			if (y < cutY)
			{
				currentNode = currentNode->NW;
			}
			else
			{
				currentNode = currentNode->SW;
			}
		}
		else
		{
			if (y < cutY)
			{
				currentNode = currentNode->NE;
			}
			else
			{
				currentNode = currentNode->SE;
			}
		}
	}
	return currentNode;
}

/**
 * Find node corresponding to pixel in quadree
 * @param pixel Pixel coordinate
 * @return Pointer to quadtree node corresponding to \a pixel
 * */
QuadTreeNode * QuadTree::find(int pixel)
{
	QuadTreeNode * currentNode = this->root;
	int x = pixel % this->width;
	int y = pixel / this->width;
	int cutX;
	int cutY;
	while (!currentNode->isLeaf())
	{
		cutX = currentNode->topLeft % this->width + currentNode->size/2;
		cutY = currentNode->topLeft / this->width + currentNode->size/2;		
		if (x < cutX)			
		{
			if (y < cutY)
			{
				currentNode = currentNode->NW;
			}
			else
			{
				currentNode = currentNode->SW;
			}
		}
		else
		{
			if (y < cutY)
			{
				currentNode = currentNode->NE;
			}
			else
			{
				currentNode = currentNode->SE;
			}
		}
	}
	return currentNode;
}

/**
 * Recursive function to calculate total number of quadtree nodes.
 * @param qtn Current quadtree node.
 * @return Recursive total number of child nodes + 1 for this node.
 * */
int memoryin(QuadTreeNode * qtn)
{
	//int a = 1;	
	if (!qtn->isLeaf())
	{
		int a = 1;
		a += memoryin(qtn->SW);
		a += memoryin(qtn->SE);
		a += memoryin(qtn->NW);
		a += memoryin(qtn->NE);
		return a;
	}
	else
	{
		return 1;
	}
	
}

/**
 * Print memory consumption information of quadtree
 * */
void QuadTree::memory()
{
	//std::cout << sizeof(QuadTreeNode) << "\n";
	//std::cout << "Quadtree nodes: " << memoryin(this->root) << "\n";
	std::cout << "Quadtree leaves: " << this->leaves->size() << "\n";
	//std::cout << this->subdivided->size() << "\n";
}


/**
 * Constructor
 * @param parent Parent \c QuadTreeNode.
 * @param level Level in quadtree.
 * @param type _SW, _SE, _NW or _NE -- 1, 2, 3 or 4 respectively.
 * @param topLeft Top left pixel of quadtree node.
 * @param size Size (width) of quadtree node.
 * @param quadTree Pointer to wrapper quadtree.
 * */
QuadTreeNode::QuadTreeNode(QuadTreeNode * parent, int level, int type, int topLeft, int size, QuadTree * quadTree)
{
	this->parent = parent;
	this->level = level;
	this->type = type;
	this->topLeft = topLeft;
	this->size = size;
	this->SW = NULL;
	this->SE = NULL;
	this->NW = NULL;
	this->NE = NULL;
	this->quadTree = quadTree;
	//this->test = new int[24];
}

/**
 * Destructor. Recursively destroys children.
 * */
QuadTreeNode::~QuadTreeNode()
{
	if(this->SW != NULL)
	{
		delete this->SW;
	}
	if(this->SE != NULL)
	{
		delete this->SE;
	}
	if(this->NW != NULL)
	{
		delete this->NW;
	}
	if(this->NE != NULL)
	{
		delete this->NE;
	}
}

/**
 * Find all adjacent nodes of same size.
 * @return Vector of all adjacent nodes.
 * */
std::vector<QuadTreeNode*> QuadTreeNode::getAllNeighbours()
{
	std::vector<QuadTreeNode*> theList;
	theList.push_back(this->getNorthNeighbour());
	theList.push_back(this->getEastNeighbour());
	theList.push_back(this->getSouthNeighbour());
	theList.push_back(this->getWestNeighbour());
	return theList;
}

/**
 * Perform a simple subdivision, only updating the leaves set, not the subdivided set or anything in respect to the A matrix.
 * @todo Probably only works when width = height
 * */
void QuadTreeNode::subdivideSimple()
{
	//this is not a leaf anymore
	this->quadTree->leaves->erase(this);
	this->quadTree->subdivided->erase(this);
	//create new nodes, add to leaves and to list of nodes which must be updated in A
	this->SW = new QuadTreeNode(this,this->level + 1, _SW, this->topLeft + this->size/2 * this->quadTree->width, this->size/2, this->quadTree);
	this->quadTree->leaves->emplace(this->SW);
	this->SE = new QuadTreeNode(this,this->level + 1, _SE, this->topLeft + this->size/2 * this->quadTree->width + this->size/2, this->size/2, this->quadTree);
	this->quadTree->leaves->emplace(this->SE);
	this->NW = new QuadTreeNode(this,this->level + 1, _NW, this->topLeft, this->size/2, this->quadTree);
	this->quadTree->leaves->emplace(this->NW);
	this->NE = new QuadTreeNode(this,this->level + 1, _NE, this->topLeft + this->size/2, this->size/2, this->quadTree);
	this->quadTree->leaves->emplace(this->NE);
}

/**
 * Subdivide quadtree node and update leaves set and subdivided set. Add neighbours to subdivided set so that the A matrix can be updated correctly.
 * */
void QuadTreeNode::subdivide()
{
	//this is not a leaf anymore
	this->quadTree->leaves->erase(this);
	this->quadTree->subdivided->erase(this);

	//create new nodes, add to leaves and to list of nodes which must be updated in A
	this->SW = new QuadTreeNode(this,this->level + 1, _SW, this->topLeft + this->size/2 * this->quadTree->width, this->size/2, this->quadTree);
	this->quadTree->leaves->emplace(this->SW);
	this->quadTree->subdivided->emplace(this->SW);
	this->SE = new QuadTreeNode(this,this->level + 1, _SE, this->topLeft + this->size/2 * this->quadTree->width + this->size/2, this->size/2, this->quadTree);
	this->quadTree->leaves->emplace(this->SE);
	this->quadTree->subdivided->emplace(this->SE);
	this->NW = new QuadTreeNode(this,this->level + 1, _NW, this->topLeft, this->size/2, this->quadTree);
	this->quadTree->leaves->emplace(this->NW);
	this->quadTree->subdivided->emplace(this->NW);
	this->NE = new QuadTreeNode(this,this->level + 1, _NE, this->topLeft + this->size/2, this->size/2, this->quadTree);
	this->quadTree->leaves->emplace(this->NE);
	this->quadTree->subdivided->emplace(this->NE);


	//we must update all neighbours in A matrix
	QuadTreeNode * neighbour = this->getNorthNeighbour();
	if(neighbour != NULL)
	{
		if(neighbour->isLeaf())
		{
			if(neighbour->level < this->level)
			{
				neighbour->subdivide();
				this->quadTree->subdivided->emplace(neighbour->SW);
				this->quadTree->subdivided->emplace(neighbour->SE);
			}
			else
			{
				this->quadTree->subdivided->emplace(neighbour);
			}
		}
		else
		{
			this->quadTree->subdivided->emplace(neighbour->SW);
			this->quadTree->subdivided->emplace(neighbour->SE);
		}
	}
	neighbour = this->getEastNeighbour();
	if(neighbour != NULL)
	{
		if(neighbour->isLeaf())
		{
			if(neighbour->level < this->level)
			{		
				neighbour->subdivide();
				this->quadTree->subdivided->emplace(neighbour->NW);
				this->quadTree->subdivided->emplace(neighbour->SW);
			}
			else
			{
				this->quadTree->subdivided->emplace(neighbour);
			}
		}
		else
		{
			this->quadTree->subdivided->emplace(neighbour->NW);
			this->quadTree->subdivided->emplace(neighbour->SW);
		}
	}
	neighbour = this->getSouthNeighbour();
	if(neighbour != NULL)
	{
		if(neighbour->isLeaf())
		{
			if(neighbour->level < this->level)
			{
				neighbour->subdivide();
				this->quadTree->subdivided->emplace(neighbour->NW);
				this->quadTree->subdivided->emplace(neighbour->NE);				
			}
			else
			{
				this->quadTree->subdivided->emplace(neighbour);
			}
		}
		else
		{
			this->quadTree->subdivided->emplace(neighbour->NW);
			this->quadTree->subdivided->emplace(neighbour->NE);
		}
	}
	neighbour = this->getWestNeighbour();
	if(neighbour != NULL)
	{
		if(neighbour->isLeaf())
		{
			if(neighbour->level < this->level)
			{
				neighbour->subdivide();
				this->quadTree->subdivided->emplace(neighbour->NE);
				this->quadTree->subdivided->emplace(neighbour->SE);
			}
			else
			{
				this->quadTree->subdivided->emplace(neighbour);
			}
		}
		else
		{
			this->quadTree->subdivided->emplace(neighbour->NE);
			this->quadTree->subdivided->emplace(neighbour->SE);
		}
	}
	//std::cout << this->topLeft + this->size/2 << "\n";
	
}

/**
 * Subdivide quadtree up to level
 * @param toLevel Level to which to subdivide
 * @todo Probably only works when width = height
 * */
void QuadTreeNode::subdivide(int toLevel)
{
	if (this->level < toLevel && this->isLeaf())
	{
		this->subdivideSimple();
		this->SW->subdivide(toLevel);
		this->SE->subdivide(toLevel);
		this->NW->subdivide(toLevel);
		this->NE->subdivide(toLevel);
	}
}

/**
 * Check whether node is leaf.
 * @return Returns \c TRUE when node is a leaf, \c FALSE otherwise.
 * */
bool QuadTreeNode::isLeaf()
{
	if (this->SW == NULL)
	{
		return true;
	}
	return false;
}

/**
 * Find northern neighbour.
 * @return Pointer to northern neighbour node. NULL if there is no neighbour.
 * */
QuadTreeNode * QuadTreeNode::getNorthNeighbour()
{
	if(this->type == _R)
	{
		return NULL;
	}
	if(this->type == _SW)
	{
		return this->parent->NW;
	}
	if(this->type == _SE)
	{
		return this->parent->NE;
	}
	QuadTreeNode * u = this->parent->getNorthNeighbour();
	if(u == NULL || u->isLeaf())
	{
		return u;
	}
	if (this->type == _NW)
	{
		return u->SW;
	}
	else
	{
		return u->SE;
	}	
}

/**
 * Find eastern neighbour.
 * @return Pointer to eastern neighbour node. NULL if there is no neighbour.
 * */
QuadTreeNode * QuadTreeNode::getEastNeighbour()
{
	if(this->type == _R)
	{
		return NULL;
	}
	if(this->type == _NW)
	{
		return this->parent->NE;
	}
	if(this->type == _SW)
	{
		return this->parent->SE;
	}
	QuadTreeNode * u = this->parent->getEastNeighbour();
	if(u == NULL || u->isLeaf())
	{
		return u;
	}
	if (this->type == _NE)
	{
		return u->NW;
	}
	else
	{
		return u->SW;
	}	
}

/**
 * Find southern neighbour.
 * @return Pointer to southern neighbour node. NULL if there is no neighbour.
 * */
QuadTreeNode * QuadTreeNode::getSouthNeighbour()
{
	if(this->type == _R)
	{
		return NULL;
	}
	if(this->type == _NW)
	{
		return this->parent->SW;
	}
	if(this->type == _NE)
	{
		return this->parent->SE;
	}
	QuadTreeNode * u = this->parent->getSouthNeighbour();
	if(u == NULL || u->isLeaf())
	{
		return u;
	}
	if (this->type == _SW)
	{
		return u->NW;
	}
	else
	{
		return u->NE;
	}	
}

/**
 * Find western neighbour.
 * @return Pointer to western neighbour node. NULL if there is no neighbour.
 * */
QuadTreeNode * QuadTreeNode::getWestNeighbour()
{
	if(this->type == _R)
	{
		return NULL;
	}
	if(this->type == _NE)
	{
		return this->parent->NW;
	}
	if(this->type == _SE)
	{
		return this->parent->SW;
	}
	QuadTreeNode * u = this->parent->getWestNeighbour();
	if(u == NULL || u->isLeaf())
	{
		return u;
	}
	if (this->type == _NW)
	{
		return u->NE;
	}
	else
	{
		return u->SE;
	}	
}

/**
* Draw quadtee mesh
* @param image Image of size WIDTH * HEIGHT to write to. Does not clear beforehand.
* */
void QuadTreeNode::printQuadTreeNode(unsigned char * image)
{
	if((this->topLeft + this->size) % this->quadTree->width == 0)
	{
		this->drawLine(image,(this->topLeft % this->quadTree->width) * this->quadTree->resolutionMultiple, (this->topLeft / this->quadTree->width * this->quadTree->resolutionMultiple),
			((this->topLeft + this->size - 1) % this->quadTree->width) * this->quadTree->resolutionMultiple, ((this->topLeft) / this->quadTree->width * this->quadTree->resolutionMultiple));
		this->drawLine(image,((this->topLeft + this->size - 1) % this->quadTree->width) * this->quadTree->resolutionMultiple, ((this->topLeft) / this->quadTree->width) * this->quadTree->resolutionMultiple,
			((this->topLeft + this->size - 1) % this->quadTree->width) * this->quadTree->resolutionMultiple, ((this->topLeft + this->size * this->quadTree->width) / this->quadTree->width) * this->quadTree->resolutionMultiple);
		this->drawLine(image,((this->topLeft + this->size - 1) % this->quadTree->width) * this->quadTree->resolutionMultiple, ((this->topLeft + this->size * this->quadTree->width) / this->quadTree->width) * this->quadTree->resolutionMultiple,
			((this->topLeft + this->size * this->quadTree->width) % this->quadTree->width) * this->quadTree->resolutionMultiple, ((this->topLeft + this->size * this->quadTree->width) / this->quadTree->width) * this->quadTree->resolutionMultiple);
		this->drawLine(image,((this->topLeft + this->size * this->quadTree->width) % this->quadTree->width) * this->quadTree->resolutionMultiple, ((this->topLeft + this->size * this->quadTree->width) / this->quadTree->width) * this->quadTree->resolutionMultiple,
			(this->topLeft % this->quadTree->width) * this->quadTree->resolutionMultiple, (this->topLeft / this->quadTree->width) * this->quadTree->resolutionMultiple);
	}
	else
	{
		this->drawLine(image,(this->topLeft % this->quadTree->width) * this->quadTree->resolutionMultiple, (this->topLeft / this->quadTree->width * this->quadTree->resolutionMultiple),
			((this->topLeft + this->size) % this->quadTree->width) * this->quadTree->resolutionMultiple, ((this->topLeft + this->size) / this->quadTree->width * this->quadTree->resolutionMultiple));
		this->drawLine(image,((this->topLeft + this->size) % this->quadTree->width) * this->quadTree->resolutionMultiple, ((this->topLeft + this->size) / this->quadTree->width) * this->quadTree->resolutionMultiple,
			((this->topLeft + this->size + this->size * this->quadTree->width) % this->quadTree->width) * this->quadTree->resolutionMultiple, ((this->topLeft + this->size + this->size * this->quadTree->width) / this->quadTree->width) * this->quadTree->resolutionMultiple);
		this->drawLine(image,((this->topLeft + this->size + this->size * this->quadTree->width) % this->quadTree->width) * this->quadTree->resolutionMultiple, ((this->topLeft + this->size + this->size * this->quadTree->width) / this->quadTree->width) * this->quadTree->resolutionMultiple,
			((this->topLeft + this->size * this->quadTree->width) % this->quadTree->width) * this->quadTree->resolutionMultiple, ((this->topLeft + this->size * this->quadTree->width) / this->quadTree->width) * this->quadTree->resolutionMultiple);
		this->drawLine(image,((this->topLeft + this->size * this->quadTree->width) % this->quadTree->width) * this->quadTree->resolutionMultiple, ((this->topLeft + this->size * this->quadTree->width) / this->quadTree->width) * this->quadTree->resolutionMultiple,
			(this->topLeft % this->quadTree->width) * this->quadTree->resolutionMultiple, (this->topLeft / this->quadTree->width) * this->quadTree->resolutionMultiple);
	}
	
}

/**
* Bresenham algorithm for vertical and horizontal lines to draw segments of quadtree mesh
* @param image Image (array) to draw to.
* @param ax Line starting point x coordinate.
* @param ay Line starting point y coordinate.
* @param bx Line endpoint x coordinate.
* @param by Line endpoint y coordinate.
*/
void QuadTreeNode::drawLine(unsigned char * image, int ax, int ay, int bx, int by)
{
	int luminance = 255;
	int rWIDTH = this->quadTree->width * this->quadTree->resolutionMultiple;
	int rHEIGHT = this->quadTree->height * this->quadTree->resolutionMultiple;

	// vertical lines
	int deltaX = ax - bx;
	int deltaY = ay - by;

	float error = 0;
	float deltaErr = std::abs((float)deltaY / deltaX);

	if (deltaX == 0) {
		int Y0, Y1;
		if (by > ay) {
			Y1 = by;
			Y0 = ay;
		}
		else {
			Y1 = ay;
			Y0 = by;
		}

		for (int y = Y0; y <= Y1; y++) {
			if(ax+rWIDTH*y > 0 && ax+rWIDTH*y < this->quadTree->width * this->quadTree->height * this->quadTree->resolutionMultiple * this->quadTree->resolutionMultiple)
			{
				image[(int)(ax+rWIDTH*y)] = 255;
				//this->imageFP->image[(int)(ax+rWIDTH*y)] = luminance;
			}
			else
			{
				image[0] = 255;
				//this->imageFP->image[0] = luminance;
			}
		}
	}
	else if (deltaY == 0)
	{
		int X0, X1;
		if (bx > ax)
		{
			X1 = bx;
			X0 = ax;
		}
		else
		{
			X1 = ax;
			X0 = bx;
		}
		for( int x = X0; x <= X1; ++x)
		{
			if(ay * rWIDTH + x > 0 && ay * rWIDTH + x < this->quadTree->width * this->quadTree->height * this->quadTree->resolutionMultiple * this->quadTree->resolutionMultiple)
			{
				image[(int)(ay * rWIDTH + x)] = 255;
			}
		}
	}
}