/*
 * P E R U N -- a physically based lightning generator
 * Master's thesis
 * Department of Computer Graphics and Interaction
 * Faculty of Electrical Engineering
 * Czech Technical University in Prague.
 *
 * Supervisor: Ing. Jaroslav Sloup
 *
 * Distributed under the GNU General Public License v3.0
 * */

/**
 * @file
 * @author Richard Dobrichovsky <dobrichovsky@gmail.com>
 * @date May 2016
 * 
 * @brief Declarations for QuadTree and QuadTreeNode classes.
 *
 * Declarations for QuadTree and QuadTreeNode classes implementing a quadtree data structure used in the adaptive grid
 * versions of conjugate gradients solvers.
 *
 * */

///Root
#define _R 0
///Southwest child
#define _SW 1
///Southeast child
#define _SE 2
///Northwest child
#define _NW 3
///Northeast child
#define _NE 4

///Direction North
#define NORTH 8
///Direction East
#define EAST 6
///Direction South
#define SOUTH 2
///Direction West
#define WEST 4

#include <vector>
#include <set>
#include <iostream>
#include <cmath>

class QuadTree;
class QuadTreeNode;

/**
 * A wrapper class to hold information about the quatree itself with a pointer to the root node.
 * @brief Class to hold information about the quatree itself with a pointer to the root node.
 */
class QuadTree
{
public:
	QuadTree(int width, int height, int minLevel, int resolutionMultiple);
	~QuadTree();	
	QuadTreeNode * insert(int pixel);
	QuadTreeNode * find(int pixel);
	void printQuadTree(unsigned char * image);
	void memory();
	/// Quadtree root
	QuadTreeNode * root;
	///Set of quadtree leaves.
	std::set<QuadTreeNode *> * leaves;
	///Set of newly subdivided quadtree nodes and their neighbours to update respective A matrix rows.
	std::set<QuadTreeNode *> * subdivided;
	/// Simulation grid / quadtree width
	int width;
	/// Simulation grid / quadtree height
	int height;
	/// Resolution multiplicator
	int resolutionMultiple;
	/// Minimum quadtree node level
	int minLevel;
	
};

/**
 * Actual quadtree node.
 * @brief Actual quadtree node.
 * */
class QuadTreeNode
{
public:
	QuadTreeNode(QuadTreeNode * parent, int level, int type, int topLeft, int size, QuadTree * quadtree);
	~QuadTreeNode();
	std::vector<QuadTreeNode*> getAllNeighbours();
	void subdivideSimple();
	void subdivide();
	void subdivide(int toLevel);
	bool isLeaf();
	QuadTreeNode * getNorthNeighbour();
	QuadTreeNode * getEastNeighbour();
	QuadTreeNode * getSouthNeighbour();
	QuadTreeNode * getWestNeighbour();
	void printQuadTreeNode(unsigned char * image);
	void drawLine(unsigned char * image, int ax, int ay, int bx, int by);
	/// Node type: _R, _SW, _SE, _NW or _NE: 0, 1, 2, 3 or 4 respectively.
	int type;
	/// Level of node in hierarchy.
	int level;
	/// Coordinates of top left pixel contained in quadtree node.
	int topLeft;
	/// Width of quadtree node.
	int size;
	/// Pointer to wrapper quadtree.
	QuadTree * quadTree;
	/// Pointer to parent. NULL if node is root.
	QuadTreeNode * parent;
	/// Pointer to southwest child. NULL if node is leaf.
	QuadTreeNode * SW;
	/// Pointer to southeast child. NULL if node is leaf.
	QuadTreeNode * SE;
	/// Pointer to northwest child. NULL if node is leaf.
	QuadTreeNode * NW;
	/// Pointer to northeast child. NULL if node is leaf.
	QuadTreeNode * NE;
};